"""
Example showing how to draw text to the screen.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.drawing_text
"""
import arcade

WINDOW_WIDTH = 1200
WINDOW_HEIGHT = 800
WINDOW_TITLE = "Drawing Text Example"
DEFAULT_LINE_HEIGHT = 45
DEFAULT_FONT_SIZE = 20

# Load fonts bundled with Arcade such as the Kenney fonts
arcade.resources.load_kenney_fonts()
arcade.resources.load_liberation_fonts()


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        super().__init__()

        self.background_color = arcade.color.BEIGE
        self.text_angle = 0
        self.time_elapsed = 0.0

    def on_update(self, delta_time):
        self.text_angle += 1
        self.time_elapsed += delta_time

    def on_draw(self):
        """
        Render the screen.
        """

        # This command should happen before we start drawing. It will clear
        # the screen to the background color, and erase what we drew last frame.
        self.clear()

        # Add the screen title
        start_x = 0
        start_y = WINDOW_HEIGHT - DEFAULT_LINE_HEIGHT * 1.5
        arcade.draw_text("Text Drawing Examples",
                         start_x,
                         start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE * 2,
                         width=WINDOW_WIDTH,
                         align="center")

        # start_x and start_y make the start point for the text. We draw a dot to make it
        # easy to see the text in relation to its start x and y.
        start_x = 10
        start_y = WINDOW_HEIGHT - DEFAULT_LINE_HEIGHT * 3
        arcade.draw_text("Fonts:",
                         start_x,
                         start_y,
                         arcade.color.FRENCH_WINE,
                         DEFAULT_FONT_SIZE, bold=True)

        # Move the y value down to create another line of text
        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_text("Default Font (Arial)",
                         start_x,
                         start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE)

        # Show some built-in fonts
        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_text("Kenney Blocks Font",
                         start_x,
                         start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         font_name="Kenney Blocks")

        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_text("Kenney Future Font",
                         start_x,
                         start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         font_name="Kenney Future")

        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_text("Kenney High Font",
                         start_x,
                         start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         font_name="Kenney High")

        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_text("Kenney High Square Font",
                         start_x,
                         start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         font_name="Kenney High Square")

        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_text("Kenney Mini Square Font",
                         start_x, start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         font_name="Kenney Mini Square")

        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_text("Kenney Pixel Font",
                         start_x, start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         font_name="Kenney Pixel")

        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_text("Kenney Pixel Square Font",
                         start_x, start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         font_name="Kenney Pixel Square")

        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_text("Kenney Rocket Font",
                         start_x, start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         font_name="Kenney Rocket")

        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_text("Kenney Rocket Square Font",
                         start_x, start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         font_name="Kenney Rocket Square")

        start_y -= DEFAULT_LINE_HEIGHT
        # When trying to use system fonts, it can be risky to specify
        # only a single font because someone else's computer might not
        # have it installed. This is especially true if they run a
        # different operating system. For example, if you are on Windows
        # and a friend has a mac or Linux, they might not have the same
        # fonts. Your game could look different or broken on their computer.
        # One way around that is to provide multiple options for draw_text
        # to try. It will use the first one it finds, and use Arial as a
        # default if it can't find any of them.
        # In the example below, draw_text is given a tuple of names for very
        # similar fonts, each of which is common on a different major
        # operating systems.
        arcade.draw_text("Times New Roman (Or closest match on system)",
                         start_x, start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         font_name=(
                             "Times New Roman",  # Comes with Windows
                             "Times",  # MacOS may sometimes have this variant
                             # Common on Linux systems + we ship it with Arcade
                             "Liberation Serif"
                         ))

        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_text("Multi-Line\ntext using\n\\n characters.",
                         start_x, start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE / 2,
                         multiline=True,
                         width=300)

        start_y -= DEFAULT_LINE_HEIGHT * 1.5
        arcade.draw_text("Wrapping really long text automatically to a new line. "
                         "The quick brown fox jumped over the lazy dogs.",
                         start_x,
                         start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE / 2,
                         multiline=True,
                         width=300)

        # --- Column 2 ---
        start_x = 750
        start_y = WINDOW_HEIGHT - DEFAULT_LINE_HEIGHT * 3
        arcade.draw_text("Text Positioning:",
                         start_x,
                         start_y,
                         arcade.color.FRENCH_WINE,
                         DEFAULT_FONT_SIZE,
                         bold=True)

        # start_x and start_y make the start point for the text.
        # We draw a dot to make it easy too see the text in relation to
        # its start x and y.
        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_point(start_x, start_y, arcade.color.BARN_RED, 5)
        arcade.draw_text("Default of 'baseline' and 'Left'",
                         start_x,
                         start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE)

        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_point(start_x, start_y, arcade.color.BARN_RED, 5)
        arcade.draw_text("'bottom' and 'left'",
                         start_x,
                         start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         anchor_x="left",
                         anchor_y="bottom")

        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_point(start_x, start_y, arcade.color.BARN_RED, 5)
        arcade.draw_text("'top' and 'left'",
                         start_x, start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         anchor_x="left",
                         anchor_y="top")

        start_y -= DEFAULT_LINE_HEIGHT * 2
        arcade.draw_point(start_x, start_y, arcade.color.BARN_RED, 5)
        arcade.draw_text("'baseline' and 'center'",
                         start_x, start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         anchor_x="center",
                         anchor_y="baseline")

        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_point(start_x, start_y, arcade.color.BARN_RED, 5)
        arcade.draw_text("'baseline' and 'right'",
                         start_x,
                         start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         anchor_x="right",
                         anchor_y="baseline")

        start_y -= DEFAULT_LINE_HEIGHT
        arcade.draw_point(start_x, start_y, arcade.color.BARN_RED, 5)
        arcade.draw_text("'center' and 'center'",
                         start_x,
                         start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         anchor_x="center",
                         anchor_y="center")

        start_y -= DEFAULT_LINE_HEIGHT * 4
        # start_x = 0
        # start_y = 0
        arcade.draw_point(start_x, start_y, arcade.color.BARN_RED, 5)
        arcade.draw_text("Rotating Text",
                         start_x, start_y,
                         arcade.color.BLACK,
                         DEFAULT_FONT_SIZE,
                         anchor_x="center",
                         anchor_y="center",
                         rotation=self.text_angle)

    def on_key_press(self, symbol: int, modifiers: int):
        """ Handle key press events """
        if symbol == arcade.key.ESCAPE:
            self.window.close()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
This module provides functionality to manage Sprites in a list
and efficiently batch drawing them. Drawing sprites using
SpriteList is orders of magnitude faster then drawing
individual sprites.
"""

from __future__ import annotations

import random
from abc import abstractmethod
from array import array
from collections import deque
from collections.abc import Callable, Collection, Iterable, Iterator, Sized
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    cast,
)

from arcade import Sprite, SpriteType, SpriteType_co, get_window, gl
from arcade.gl import Program, Texture2D
from arcade.gl.buffer import Buffer
from arcade.gl.types import BlendFunction, OpenGlFilter, PyGLenum
from arcade.gl.vertex_array import Geometry
from arcade.types import RGBA255, Color, Point2, RGBANormalized, RGBOrA255, RGBOrANormalized
from arcade.utils import copy_dunders_unimplemented

if TYPE_CHECKING:
    from arcade import DefaultTextureAtlas, Texture
    from arcade.texture_atlas import TextureAtlasBase

# The default capacity from spritelists
_DEFAULT_CAPACITY = 100


class SpriteSequence(Collection[SpriteType_co]):
    """A read-only view of a :py:class:`.SpriteList`.

    Like other read-only generics such as :py:class:`collections.abc.Sequence`,
    a `SpriteSequence` requires sprites be of a covariant type relative to their
    annotated type.

    See :py:class:`.SpriteList` for more details.
    """

    from ..sprite_list import spatial_hash as sh

    @property
    @abstractmethod
    def spatial_hash(self) -> sh.ReadOnlySpatialHash[SpriteType_co] | None: ...

    @abstractmethod
    def __getitem__(self, index: int) -> SpriteType_co:
        """Return the sprite at the given index."""
        ...

    @abstractmethod
    def update(self, delta_time: float = 1 / 60, *args, **kwargs) -> None:
        """
        Call the update() method on each sprite in the list.

        Args:
            delta_time: Time since last update in seconds
            *args: Additional positional arguments
            **kwargs: Additional keyword arguments
        """
        ...

    @abstractmethod
    def update_animation(self, delta_time: float = 1 / 60, *args, **kwargs) -> None:
        """
        Call the update_animation in every sprite in the sprite list.

        Args:
            delta_time: Time since last update in seconds
            *args: Additional positional arguments
            **kwargs: Additional keyword arguments
        """
        ...

    @abstractmethod
    def draw(
        self,
        *,
        filter: PyGLenum | OpenGlFilter | None = None,
        pixelated: bool | None = None,
        blend_function: BlendFunction | None = None,
    ) -> None:
        """
        Draw this list of sprites.

        Uninitialized sprite lists will first create OpenGL resources
        before drawing. This may cause a performance stutter when the
        following are true:

        1. You created the sprite list with ``lazy=True``
        2. You did not call :py:meth:`~SpriteList.initialize` before drawing
        3. You are initializing many sprites and/or lists at once

        See :ref:`pg_spritelist_advanced_lazy_spritelists` to learn more.

        Args:
            filter:
                Optional parameter to set OpenGL filter, such as
                `gl.GL_NEAREST` to avoid smoothing.
            pixelated:
                ``True`` for pixelated and ``False`` for smooth interpolation.
                Shortcut for setting filter to GL_NEAREST for a pixelated look.
                The filter parameter have precedence over this.
            blend_function:
                Optional parameter to set the OpenGL blend function used for drawing
                the sprite list, such as 'arcade.Window.ctx.BLEND_ADDITIVE' or
                'arcade.Window.ctx.BLEND_DEFAULT'
        """
        ...

    @abstractmethod
    def draw_hit_boxes(
        self, color: RGBOrA255 = (0, 0, 0, 255), line_thickness: float = 1.0
    ) -> None:
        """
        Draw all the hit boxes in this list.

        .. warning:: This method is slow and should only be used for debugging.

        Args:
            color: The color of the hit boxes
            line_thickness: The thickness of the lines
        """
        ...

    @abstractmethod
    def _write_sprite_buffers_to_gpu(self) -> None: ...


@copy_dunders_unimplemented  # Temp fixes https://github.com/pythonarcade/arcade/issues/2074
class SpriteList(SpriteSequence[SpriteType]):
    """
    The purpose of the spriteList is to batch draw a list of sprites.
    Drawing single sprites will not get you anywhere performance wise
    as the number of sprites in your project increases. The spritelist
    contains many low level optimizations taking advantage of your
    graphics processor. To put things into perspective, a spritelist
    can contain tens of thousands of sprites without any issues.
    Sprites outside the viewport/window will not be rendered.

    If the spritelist are going to be used for collision it's a good
    idea to enable spatial hashing. Especially if no sprites are moving.
    This will make collision checking **a lot** faster.
    In technical terms collision checking is ``O(1)`` with spatial hashing
    enabled and ``O(N)`` without. However, if you have a
    list of moving sprites the cost of updating the spatial hash
    when they are moved can be greater than what you save with
    spatial collision checks. This needs to be profiled on a
    case by case basis.

    For the advanced options check the advanced section in the
    Arcade documentation.

    Args:
        use_spatial_hash:
            If set to True, this will make creating a sprite, and moving a sprite
            in the SpriteList slower, but it will speed up collision detection
            with items in the SpriteList. Great for doing collision detection
            with static walls/platforms in large maps.
        spatial_hash_cell_size:
            The cell size of the spatial hash (default: 128)
        atlas:
            (Advanced) The texture atlas for this sprite list. If no
            atlas is supplied the global/default one will be used.
        capacity:
            (Advanced) The initial capacity of the internal buffer.
            It's a suggestion for the maximum amount of sprites this list
            can hold. Can normally be left with default value.
        lazy:
            (Advanced) ``True`` delays creating OpenGL resources
            for the sprite list until either its :py:meth:`~SpriteList.draw`
            or :py:meth:`~SpriteList.initialize` method is called. See
            :ref:`pg_spritelist_advanced_lazy_spritelists` to learn more.
        visible:
            Setting this to False will cause the SpriteList to not
            be drawn. When draw is called, the method will just return without drawing.
    """

    #: The default texture filter used when no other filter is specified.
    #: This can be used to change the global default for all spritelists
    #:
    #: Example::
    #:
    #:     from arcade import gl
    #:     # Set global default to nearest filtering (pixelated)
    #:     arcade.SpriteList.DEFAULT_TEXTURE_FILTER = gl.NEAREST, gl.NEAREST
    #:     # Set global default to linear filtering (smooth). This is the default.
    #:     arcade.SpriteList.DEFAULT_TEXTURE_FILTER = gl.NEAREST, gl.NEAREST
    DEFAULT_TEXTURE_FILTER: ClassVar[tuple[int, int]] = gl.LINEAR, gl.LINEAR

    # Declare `special_hash` as an attribute that implements the abstract
    # property from `SpriteSequence`. It needs an explicit type here because
    # it is better than the inherited type.
    # More subtle: it requires to be initialized as a *class* attribute with
    # `= None` to "delete" the abstract property definition from the class.
    # Without that trick, attempt to instantiate a SpriteList results in a
    #   TypeError: Can't instantiate abstract class SpriteList
    #   without an implementation for abstract method 'spatial_hash'
    # The abstract property is actually implemented as an attribute (for
    # efficiency), so it is OK to silence the issue like that.
    from ..sprite_list import spatial_hash as sh

    spatial_hash: sh.SpatialHash[SpriteType] | None = None

    def __init__(
        self,
        use_spatial_hash: bool = False,
        spatial_hash_cell_size: int = 128,
        atlas: TextureAtlasBase | None = None,
        capacity: int = 100,
        lazy: bool = False,
        visible: bool = True,
    ) -> None:
        self.program: Program | None = None
        self._atlas: TextureAtlasBase | None = atlas
        self._initialized = False
        self._lazy = lazy
        self._visible = visible
        self._blend = True
        self._color: tuple[float, float, float, float] = (1.0, 1.0, 1.0, 1.0)

        # The initial capacity of the spritelist buffers (internal)
        self._buf_capacity = abs(capacity) or _DEFAULT_CAPACITY
        # The initial capacity of the index buffer (internal)
        self._idx_capacity = abs(capacity) or _DEFAULT_CAPACITY
        # The number of slots used in the sprite buffer
        self._sprite_buffer_slots = 0
        # Number of slots used in the index buffer
        self._sprite_index_slots = 0
        # List of free slots in the sprite buffers. These are filled when sprites are removed.
        self._sprite_buffer_free_slots: deque[int] = deque()

        # List of sprites in the sprite list
        self.sprite_list: list[SpriteType] = []
        # Buffer slots for the sprites (excluding index buffer)
        # This has nothing to do with the index in the spritelist itself
        self.sprite_slot: dict[SpriteType, int] = dict()

        # Python representation of buffer data
        self._sprite_pos_data = array("f", [0] * self._buf_capacity * 3)
        self._sprite_size_data = array("f", [0] * self._buf_capacity * 2)
        self._sprite_angle_data = array("f", [0] * self._buf_capacity)
        self._sprite_color_data = array("B", [0] * self._buf_capacity * 4)
        self._sprite_texture_data = array("f", [0] * self._buf_capacity)
        # Index buffer
        self._sprite_index_data = array("i", [0] * self._idx_capacity)

        # Define and annotate storage space for buffers
        self._sprite_pos_buf: Buffer | None = None
        self._sprite_size_buf: Buffer | None = None
        self._sprite_angle_buf: Buffer | None = None
        self._sprite_color_buf: Buffer | None = None
        self._sprite_texture_buf: Buffer | None = None

        # Index buffer
        self._sprite_index_buf: Buffer | None = None

        self._geometry: Geometry | None = None

        # Flags for signaling if a buffer needs to be written to the OpenGL buffer
        self._sprite_pos_changed: bool = False
        self._sprite_size_changed: bool = False
        self._sprite_angle_changed: bool = False
        self._sprite_color_changed: bool = False
        self._sprite_texture_changed: bool = False
        self._sprite_index_changed: bool = False

        # Used in collision detection optimization
        from .spatial_hash import SpatialHash

        self._spatial_hash_cell_size = spatial_hash_cell_size
        self.spatial_hash = None
        if use_spatial_hash:
            self.spatial_hash = SpatialHash(cell_size=self._spatial_hash_cell_size)

        self.properties: dict[str, Any] | None = None

        # Check if the window/context is available
        try:
            get_window()
            if not self._lazy:
                self._init_deferred()
        except RuntimeError:
            pass

    def _init_deferred(self) -> None:
        """
        Since spritelist can be created before the window we need to defer initialization.

        It also makes us able to support lazy loading.
        """
        if self._initialized:
            return

        self.ctx = get_window().ctx
        self.program = self.ctx.sprite_list_program_cull
        if not self._atlas:
            self._atlas = self.ctx.default_atlas

        # Buffers for each sprite attribute (read by shader) with initial capacity
        self._sprite_pos_buf = self.ctx.buffer(reserve=self._buf_capacity * 12)  # 3 x 32 bit floats
        self._sprite_size_buf = self.ctx.buffer(reserve=self._buf_capacity * 8)  # 2 x 32 bit floats
        self._sprite_angle_buf = self.ctx.buffer(reserve=self._buf_capacity * 4)  # 32 bit float
        self._sprite_color_buf = self.ctx.buffer(reserve=self._buf_capacity * 4)  # 4 x bytes colors
        self._sprite_texture_buf = self.ctx.buffer(reserve=self._buf_capacity * 4)  # 32 bit int
        # Index buffer
        self._sprite_index_buf = self.ctx.buffer(
            reserve=self._idx_capacity * 4
        )  # 32 bit unsigned integers

        contents = [
            gl.BufferDescription(self._sprite_pos_buf, "3f", ["in_pos"]),
            gl.BufferDescription(self._sprite_size_buf, "2f", ["in_size"]),
            gl.BufferDescription(self._sprite_angle_buf, "1f", ["in_angle"]),
            gl.BufferDescription(self._sprite_texture_buf, "1f", ["in_texture"]),
            gl.BufferDescription(
                self._sprite_color_buf,
                "4f1",
                ["in_color"],
            ),
        ]
        self._geometry = self.ctx.geometry(
            contents,
            index_buffer=self._sprite_index_buf,
            index_element_size=4,  # 32 bit integers
        )

        self._initialized = True

        # Load all the textures and write texture coordinates into buffers.
        for sprite in self.sprite_list:
            if sprite._texture is None:
                raise ValueError("Attempting to use a sprite without a texture")
            self._update_texture(sprite)
            if hasattr(sprite, "textures"):
                if TYPE_CHECKING:
                    assert isinstance(sprite, Sprite)
                for texture in sprite.textures or []:
                    self._atlas.add(texture)

        self._sprite_pos_changed = True
        self._sprite_size_changed = True
        self._sprite_angle_changed = True
        self._sprite_color_changed = True
        self._sprite_texture_changed = True
        self._sprite_index_changed = True

    def __len__(self) -> int:
        """Return the length of the sprite list."""
        return len(self.sprite_list)

    def __contains__(self, sprite: object) -> bool:
        """Return if the sprite list contains the given sprite"""
        return sprite in self.sprite_slot

    def __iter__(self) -> Iterator[SpriteType]:
        """Return an iterable object of sprites."""
        return iter(self.sprite_list)

    def __getitem__(self, i: int) -> SpriteType:
        return self.sprite_list[i]

    def __setitem__(self, index: int, sprite: SpriteType) -> None:
        """Replace a sprite at a specific index"""
        try:
            existing_index = self.sprite_list.index(sprite)  # raise ValueError
            if existing_index == index:
                return
            raise Exception(f"Sprite is already in the list (index {existing_index})")
        except ValueError:
            pass

        sprite_to_be_removed = self.sprite_list[index]
        sprite_to_be_removed._unregister_sprite_list(self)
        self.sprite_list[index] = sprite  # Replace sprite
        sprite.register_sprite_list(self)

        if self.spatial_hash is not None:
            self.spatial_hash.remove(sprite_to_be_removed)
            self.spatial_hash.add(sprite)

        # Steal the slot from the old sprite
        slot = self.sprite_slot[sprite_to_be_removed]
        del self.sprite_slot[sprite_to_be_removed]
        self.sprite_slot[sprite] = slot

        # Update the internal sprite buffer data
        self._update_all(sprite)

    @property
    def visible(self) -> bool:
        """
        Get or set the visible flag for this spritelist.

        If visible is ``False`` the ``draw()`` has no effect.
        """
        return self._visible

    @visible.setter
    def visible(self, value: bool) -> None:
        self._visible = value

    @property
    def blend(self) -> bool:
        """Enable or disable alpha blending for the spritelist."""
        return self._blend

    @blend.setter
    def blend(self, value: bool) -> None:
        self._blend = value

    @property
    def color(self) -> Color:
        """
        Get or set the multiply color for all sprites in the list RGBA integers

        This will affect all sprites in the list, and each value must be
        between 0 and 255.

        The color may be specified as any of the following:

        * an RGBA :py:class:`tuple` with each channel value between 0 and 255
        * an instance of :py:class:`~arcade.types.Color`
        * an RGB :py:class:`tuple`, in which case the color will be treated as opaque

        Each individual sprite can also be assigned a color via its
        :py:attr:`~arcade.BasicSprite.color` property.

        When :py:meth:`.SpriteList.draw` is called, each pixel will default
        to a value equivalent to the following:

        1. Convert the sampled texture, sprite, and list colors into normalized floats (0.0 to 1.0)
        2. Multiply the color channels together: ``texture_color * sprite_color * spritelist_color``
        3. Multiply the floating point values by 255 and round the result
        """
        return Color.from_normalized(self._color)

    @color.setter
    def color(self, color: RGBA255) -> None:
        self._color = Color.from_iterable(color).normalized

    @property
    def color_normalized(self) -> RGBANormalized:
        """
        Get or set the spritelist color in normalized form (0.0 -> 1.0 floats).

        This property works the same as :py:attr:`~arcade.SpriteList.color`.
        """
        return self._color

    @color_normalized.setter
    def color_normalized(self, value: RGBOrANormalized) -> None:
        try:
            r, g, b, *_a = value
            assert len(_a) <= 1
        except (ValueError, AssertionError) as e:
            raise ValueError("color_normalized must unpack as 3 or 4 float values") from e

        self._color = r, g, b, _a[0] if _a else 1.0

    @property
    def alpha(self) -> int:
        """
        Get or set the alpha/transparency of the entire spritelist.

        This is a byte value from 0 to 255 were 0 is completely
        transparent/invisible and 255 is opaque.
        """
        return int(self._color[3] * 255)

    @alpha.setter
    def alpha(self, value: int) -> None:
        self._color = self._color[0], self._color[1], self._color[2], value / 255

    @property
    def alpha_normalized(self) -> float:
        """
        Get or set the alpha/transparency of all the sprites in the list.

        This is a floating point number from 0.0 to 1.0 were 0.0 is completely
        transparent/invisible and 1.0 is opaque.

        This is a shortcut for setting the alpha value in the spritelist color.
        """
        return self._color[3]

    @alpha_normalized.setter
    def alpha_normalized(self, value: float) -> None:
        self._color = self._color[0], self._color[1], self._color[2], value

    @property
    def atlas(self) -> TextureAtlasBase | None:
        """Get the texture atlas for this sprite list"""
        return self._atlas

    @property
    def geometry(self) -> Geometry:
        """
        Returns the internal OpenGL geometry for this spritelist.
        This can be used to execute custom shaders with the
        spritelist data.

        One or multiple of the following inputs must be defined in your vertex shader::

            in vec2 in_pos;
            in float in_angle;
            in vec2 in_size;
            in float in_texture;
            in vec4 in_color;
        """
        if not self._initialized:
            self.initialize()

        return self._geometry  # type: ignore

    @property
    def buffer_positions(self) -> Buffer:
        """
        Get the internal OpenGL position buffer for this spritelist.

        The buffer contains 32 bit float values with
        x, y and z positions. These are the center positions
        for each sprite.

        This buffer is attached to the :py:attr:`~arcade.SpriteList.geometry`
        instance with name ``in_pos``.
        """
        if self._sprite_pos_buf is None:
            raise ValueError("SpriteList is not initialized")
        return self._sprite_pos_buf

    @property
    def buffer_sizes(self) -> Buffer:
        """
        Get the internal OpenGL size buffer for this spritelist.

        The buffer contains 32 bit float width and height values.

        This buffer is attached to the :py:attr:`~arcade.SpriteList.geometry`
        instance with name ``in_size``.
        """
        if self._sprite_size_buf is None:
            raise ValueError("SpriteList is not initialized")
        return self._sprite_size_buf

    @property
    def buffer_angles(self) -> Buffer:
        """
        Get the internal OpenGL angle buffer for the spritelist.

        This buffer contains a series of 32 bit floats
        representing the rotation angle for each sprite in degrees.

        This buffer is attached to the :py:attr:`~arcade.SpriteList.geometry`
        instance with name ``in_angle``.
        """
        if self._sprite_angle_buf is None:
            raise ValueError("SpriteList is not initialized")
        return self._sprite_angle_buf

    @property
    def buffer_colors(self) -> Buffer:
        """
        Get the internal OpenGL color buffer for this spritelist.

        This buffer contains a series of 32 bit floats representing
        the RGBA color for each sprite. 4 x floats = RGBA.


        This buffer is attached to the :py:attr:`~arcade.SpriteList.geometry`
        instance with name ``in_color``.
        """
        if self._sprite_color_buf is None:
            raise ValueError("SpriteList is not initialized")
        return self._sprite_color_buf

    @property
    def buffer_textures(self) -> Buffer:
        """
        Get the internal openGL texture id buffer for the spritelist.

        This buffer contains a series of single 32 bit floats referencing
        a texture ID. This ID references a texture in the texture
        atlas assigned to this spritelist. The ID is used to look up
        texture coordinates in a 32bit floating point texture the
        texture atlas provides. This system makes sure we can resize
        and rebuild a texture atlas without having to rebuild every
        single spritelist.

        This buffer is attached to the :py:attr:`~arcade.SpriteList.geometry`
        instance with name ``in_texture``.

        Note that it should ideally an unsigned integer, but due to
        compatibility we store them as 32 bit floats. We cast them
        to integers in the shader.
        """
        if self._sprite_texture_buf is None:
            raise ValueError("SpriteList is not initialized")
        return self._sprite_texture_buf

    @property
    def buffer_indices(self) -> Buffer:
        """
        Get the internal index buffer for this spritelist.

        The data in the other buffers are not in the correct order
        matching ``spritelist[i]``. The index buffer has to be
        used used to resolve the right order. It simply contains
        a series of integers referencing locations in the other buffers.

        Also note that the length of this buffer might be bigger than
        the number of sprites. Rely on ``len(spritelist)`` for the
        correct length.

        This index buffer is attached to the :py:attr:`~arcade.SpriteList.geometry`
        instance and will be automatically be applied the the input buffers
        when rendering or transforming.
        """
        if self._sprite_index_buf is None:
            raise ValueError("SpriteList is not initialized")
        return self._sprite_index_buf

    def _next_slot(self) -> int:
        """
        Get the next available slot in sprite buffers

        :return: index slot, buffer_slot
        """
        # Reuse old slots from deleted sprites
        if self._sprite_buffer_free_slots:
            return self._sprite_buffer_free_slots.popleft()

        # Add a new slot
        buff_slot = self._sprite_buffer_slots
        self._sprite_buffer_slots += 1
        self._grow_sprite_buffers()  # We might need to increase our buffers
        return buff_slot

    def index(self, sprite: SpriteType) -> int:
        """
        Return the index of a sprite in the spritelist

        Args:
            sprite: Sprite to find and return the index of
        """
        return self.sprite_list.index(sprite)

    def clear(self, *, capacity: int | None = None, deep: bool = True) -> None:
        """
        Remove all the sprites resetting the spritelist
        to it's initial state.

        The complexity of this method is ``O(N)`` with a deep clear (default).

        If ALL the sprites in the list gets garbage collected with the list itself
        you can do an ``O(1)``` clear using ``deep=False``. **Make sure you know
        exactly what you are doing before using this option.** Any lingering sprite
        reference will cause a massive memory leak. The ``deep`` option will
        iterate all the sprites and remove their references to this spritelist.
        Sprite and SpriteList have a circular reference for performance reasons.

        Args:
            deep: Whether to do a deep clear or not. Default is ``True``.
            capacity: The size of the internal buffers used to store the sprites.
                      Defaults to preserving the current capacity.
        """
        from .spatial_hash import SpatialHash

        # Manually remove the spritelist from all sprites
        if deep:
            for sprite in self.sprite_list:
                sprite._unregister_sprite_list(self)

        self.sprite_list = []
        self.sprite_slot = dict()

        # Reset SpatialHash
        if self.spatial_hash is not None:
            self.spatial_hash = SpatialHash(cell_size=self._spatial_hash_cell_size)

        # Clear the slot_idx and slot info and other states
        capacity = abs(capacity or self._buf_capacity)

        self._buf_capacity = capacity
        self._idx_capacity = capacity
        self._sprite_buffer_slots = 0
        self._sprite_index_slots = 0
        self._sprite_buffer_free_slots = deque()

        # Reset buffers
        # Python representation of buffer data
        self._sprite_pos_data = array("f", [0] * self._buf_capacity * 3)
        self._sprite_size_data = array("f", [0] * self._buf_capacity * 2)
        self._sprite_angle_data = array("f", [0] * self._buf_capacity)
        self._sprite_color_data = array("B", [0] * self._buf_capacity * 4)
        self._sprite_texture_data = array("f", [0] * self._buf_capacity)
        # Index buffer
        self._sprite_index_data = array("I", [0] * self._idx_capacity)

        if self._initialized:
            self._initialized = False
            self._init_deferred()

    def pop(self, index: int = -1) -> SpriteType:
        """
        Attempt to pop a sprite from the list.

        This works like :external:ref:`popping from <tut-morelists>` a
        standard Python :py:class:`list`:

        #. If the list is empty, raise an :py:class:`IndexError`
        #. If no ``index`` is passed, try to pop the last
           :py:class:`Sprite` in the list

        This is the most efficient way to remove a sprite from the list.
        The complexity of this method is ``O(1)``.

        Args:
            index:
                Index of sprite to remove (defaults to ``-1`` for the last item)
        """
        if len(self.sprite_list) == 0:
            raise IndexError("pop from empty list")

        sprite = self.sprite_list.pop(index)
        try:
            slot = self.sprite_slot[sprite]
        except KeyError:
            raise ValueError("Sprite is not in the SpriteList")

        sprite._unregister_sprite_list(self)
        del self.sprite_slot[sprite]
        self._sprite_buffer_free_slots.append(slot)

        _ = self._sprite_index_data.pop(index)
        self._sprite_index_data.append(0)
        self._sprite_index_slots -= 1
        self._sprite_index_changed = True

        if self.spatial_hash is not None:
            self.spatial_hash.remove(sprite)

        return sprite

    def append(self, sprite: SpriteType) -> None:
        """
        Add a new sprite to the list.

        Args:
            sprite: Sprite to add to the list.
        """
        # print(f"{id(self)} : {id(sprite)} append")
        if sprite in self.sprite_slot:
            raise ValueError("Sprite already in SpriteList")

        slot = self._next_slot()
        self.sprite_slot[sprite] = slot
        self.sprite_list.append(sprite)
        sprite.register_sprite_list(self)

        self._update_all(sprite)

        # Add sprite to the end of the index buffer
        idx_slot = self._sprite_index_slots
        self._sprite_index_slots += 1
        self._grow_index_buffer()
        self._sprite_index_data[idx_slot] = slot
        self._sprite_index_changed = True

        if self.spatial_hash is not None:
            self.spatial_hash.add(sprite)

        if self._initialized:
            if sprite.texture is None:
                raise ValueError("Sprite must have a texture when added to a SpriteList")
            self._atlas.add(sprite.texture)  # type: ignore

    def swap(self, index_1: int, index_2: int) -> None:
        """
        Swap two sprites by index.

        Args:
            index_1: Item index to swap
            index_2: Item index to swap
        """
        # Swap order in python spritelist
        sprite_1 = self.sprite_list[index_1]
        sprite_2 = self.sprite_list[index_2]
        self.sprite_list[index_1] = sprite_2
        self.sprite_list[index_2] = sprite_1

        # Swap order in index buffer to change rendering order
        slot_1 = self.sprite_slot[sprite_1]
        slot_2 = self.sprite_slot[sprite_2]
        i1 = self._sprite_index_data.index(slot_1)
        i2 = self._sprite_index_data.index(slot_2)
        self._sprite_index_data[i1] = slot_2
        self._sprite_index_data[i2] = slot_1

        self._sprite_index_changed = True

    def remove(self, sprite: SpriteType) -> None:
        """
        Remove a specific sprite from the list.

        Note that this method is ``O(N)`` in complexity and will have
        and increased cost the more sprites you have in the list.
        A faster option is to use :py:meth:`pop` or :py:meth:`swap`.

        Args:
            sprite: Item to remove from the list
        """
        try:
            slot = self.sprite_slot[sprite]
        except KeyError:
            raise ValueError("Sprite is not in the SpriteList")

        index = self.sprite_list.index(sprite)
        self.sprite_list.pop(index)
        sprite._unregister_sprite_list(self)
        del self.sprite_slot[sprite]

        self._sprite_buffer_free_slots.append(slot)

        self._sprite_index_data.pop(index)
        self._sprite_index_data.append(0)
        self._sprite_index_slots -= 1
        self._sprite_index_changed = True

        if self.spatial_hash is not None:
            self.spatial_hash.remove(sprite)

    def extend(self, sprites: Iterable[SpriteType]) -> None:
        """
        Extends the current list with the given iterable

        Args:
            sprites: Iterable of Sprites to add to the list
        """
        for sprite in sprites:
            self.append(sprite)

    def insert(self, index: int, sprite: SpriteType) -> None:
        """
        Inserts a sprite at a given index.

        Args:
            index: The index at which to insert
            sprite: The sprite to insert
        """
        if sprite in self.sprite_list:
            raise ValueError("Sprite is already in list")

        index = max(min(len(self.sprite_list), index), 0)

        self.sprite_list.insert(index, sprite)
        sprite.register_sprite_list(self)

        # Allocate a new slot and write the data
        slot = self._next_slot()
        self.sprite_slot[sprite] = slot
        self._update_all(sprite)

        # Allocate room in the index buffer
        self._normalize_index_buffer()
        # idx_slot = self._sprite_index_slots
        self._sprite_index_slots += 1
        self._grow_index_buffer()
        self._sprite_index_data.insert(index, slot)
        self._sprite_index_data.pop()

        if self.spatial_hash is not None:
            self.spatial_hash.add(sprite)

    def reverse(self) -> None:
        """Reverses the current list in-place"""
        # Ensure the index buffer is normalized
        self._normalize_index_buffer()

        # Reverse the sprites and index buffer
        self.sprite_list.reverse()
        # This seems to be the reasonable way to reverse a subset of an array
        reverse_data = self._sprite_index_data[0 : len(self.sprite_list)]
        reverse_data.reverse()
        self._sprite_index_data[0 : len(self.sprite_list)] = reverse_data

        self._sprite_index_changed = True

    def shuffle(self) -> None:
        """Shuffles the current list in-place"""
        # The only thing we need to do when shuffling is
        # to shuffle the sprite_list and index buffer in
        # in the same operation. We don't change the sprite buffers

        # Make sure the index buffer is the same length as the sprite list
        self._normalize_index_buffer()

        # zip index and sprite into pairs and shuffle
        pairs = list(zip(self.sprite_list, self._sprite_index_data))
        random.shuffle(pairs)

        # Reconstruct the lists again from pairs
        sprites, indices = cast(tuple[list[SpriteType], list[int]], zip(*pairs))
        self.sprite_list = list(sprites)
        self._sprite_index_data = array("I", indices)

        # Resize the index buffer to the original capacity
        if len(self._sprite_index_data) < self._idx_capacity:
            extend_by = self._idx_capacity - len(self._sprite_index_data)
            self._sprite_index_data.extend([0] * extend_by)

        self._sprite_index_changed = True

    def sort(self, *, key: Callable, reverse: bool = False) -> None:
        """
        Sort the spritelist in place using ``<`` comparison between sprites.
        This function is similar to python's :py:meth:`list.sort`.

        Example sorting sprites based on y-axis position using a lambda::

            # Normal order
            spritelist.sort(key=lambda x: x.position[1])
            # Reversed order
            spritelist.sort(key=lambda x: x.position[1], reverse=True)

        Example sorting sprites using a function::

            # More complex sorting logic can be applied, but let's just stick to y position
            def create_y_pos_comparison(sprite):
                return sprite.position[1]

            spritelist.sort(key=create_y_pos_comparison)

        Args:
            key:
                A function taking a sprite as an argument returning a comparison key
            reverse:
                If set to ``True`` the sprites will be sorted in reverse
        """
        # Ensure the index buffer is normalized
        self._normalize_index_buffer()

        # In-place sort the spritelist
        self.sprite_list.sort(key=key, reverse=reverse)
        # Loop over the sorted sprites and assign new values in index buffer
        for i, sprite in enumerate(self.sprite_list):
            self._sprite_index_data[i] = self.sprite_slot[sprite]

        self._sprite_index_changed = True

    def disable_spatial_hashing(self) -> None:
        """Deletes the internal spatial hash object."""
        self.spatial_hash = None

    def enable_spatial_hashing(self, spatial_hash_cell_size: int = 128) -> None:
        """
        Turn on spatial hashing unless it is already enabled with the same cell size.

        Args:
            spatial_hash_cell_size: The size of the cell in the spatial hash.
        """
        if self.spatial_hash is None or self.spatial_hash.cell_size != spatial_hash_cell_size:
            from .spatial_hash import SpatialHash

            self.spatial_hash = SpatialHash(cell_size=spatial_hash_cell_size)
            self._recalculate_spatial_hashes()

    def _recalculate_spatial_hashes(self) -> None:
        if self.spatial_hash is None:
            from .spatial_hash import SpatialHash

            self.spatial_hash = SpatialHash(cell_size=self._spatial_hash_cell_size)

        self.spatial_hash.reset()
        for sprite in self.sprite_list:
            self.spatial_hash.add(sprite)

    def update(self, delta_time: float = 1 / 60, *args, **kwargs) -> None:
        for sprite in self.sprite_list:
            sprite.update(delta_time, *args, **kwargs)

    def update_animation(self, delta_time: float = 1 / 60, *args, **kwargs) -> None:
        for sprite in self.sprite_list:
            sprite.update_animation(delta_time, *args, **kwargs)

    def _get_center(self) -> tuple[float, float]:
        """Get the mean center coordinates of all sprites in the list."""
        x = sum(sprite.center_x for sprite in self.sprite_list) / len(self.sprite_list)
        y = sum(sprite.center_y for sprite in self.sprite_list) / len(self.sprite_list)
        return x, y

    center = property(_get_center)

    def rescale(self, factor: float) -> None:
        """Rescale all sprites in the list relative to the spritelists center."""
        for sprite in self.sprite_list:
            sprite.rescale_relative_to_point(self.center, factor)

    def move(self, change_x: float, change_y: float) -> None:
        """
        Moves all Sprites in the list by the same amount.
        This can be a very expensive operation depending on the
        size of the sprite list.

        Args:
            change_x: Amount to change all x values by
            change_y: Amount to change all y values by
        """
        for sprite in self.sprite_list:
            sprite.center_x += change_x
            sprite.center_y += change_y

    def preload_textures(self, texture_list: Iterable[Texture]) -> None:
        """
        Preload a set of textures that will be used for sprites in this
        sprite list.

        Args:
            texture_list: List of textures.
        """
        if not self.ctx:
            raise ValueError("Cannot preload textures before the window is created")

        for texture in texture_list:
            # Ugly spacing is a fast workaround for None type checking issues
            self._atlas.add(texture)  # type: ignore

    def write_sprite_buffers_to_gpu(self) -> None:
        """
        Ensure buffers are resized and fresh sprite data is written into the internal
        sprite buffers.

        This is automatically called in :py:meth:`SpriteList.draw`, but there are
        instances when using custom shaders we need to force this to happen since
        we might have not called :py:meth:`SpriteList.draw` since the spritelist
        was modified.

        If you have added, removed, moved or changed ANY sprite property this method
        will synchronize the data on the gpu side (buffer resizing and writing in
        new data).
        """
        self._write_sprite_buffers_to_gpu()

    def _write_sprite_buffers_to_gpu(self) -> None:
        if self._sprite_pos_changed and self._sprite_pos_buf:
            self._sprite_pos_buf.orphan()
            self._sprite_pos_buf.write(self._sprite_pos_data)
            self._sprite_pos_changed = False

        if self._sprite_size_changed and self._sprite_size_buf:
            self._sprite_size_buf.orphan()
            self._sprite_size_buf.write(self._sprite_size_data)
            self._sprite_size_changed = False

        if self._sprite_angle_changed and self._sprite_angle_buf:
            self._sprite_angle_buf.orphan()
            self._sprite_angle_buf.write(self._sprite_angle_data)
            self._sprite_angle_changed = False

        if self._sprite_color_changed and self._sprite_color_buf:
            self._sprite_color_buf.orphan()
            self._sprite_color_buf.write(self._sprite_color_data)
            self._sprite_color_changed = False

        if self._sprite_texture_changed and self._sprite_texture_buf:
            self._sprite_texture_buf.orphan()
            self._sprite_texture_buf.write(self._sprite_texture_data)
            self._sprite_texture_changed = False

        if self._sprite_index_changed and self._sprite_index_buf:
            self._sprite_index_buf.orphan()
            self._sprite_index_buf.write(self._sprite_index_data)
            self._sprite_index_changed = False

    def initialize(self) -> None:
        """
        Request immediate creation of OpenGL resources for this list.

        Calling this method is optional. It only has an effect for lists
        created with ``lazy=True``. If this method is not called,
        uninitialized sprite lists will automatically initialize OpenGL
        resources on their first :py:meth:`~SpriteList.draw` call instead.

        This method is useful for performance optimization, advanced
        techniques, and writing tests. Do not call it across thread
        boundaries. See :ref:`pg_spritelist_advanced_lazy_spritelists`
        to learn more.
        """
        self._init_deferred()

    def draw(
        self,
        *,
        filter: PyGLenum | OpenGlFilter | None = None,
        pixelated: bool | None = None,
        blend_function: BlendFunction | None = None,
    ) -> None:
        if len(self.sprite_list) == 0 or not self._visible or self.alpha_normalized == 0.0:
            return

        self._init_deferred()
        if not self.program:
            raise ValueError("Attempting to render without shader program.")
        self._write_sprite_buffers_to_gpu()

        prev_blend_func = self.ctx.blend_func
        if self._blend:
            self.ctx.enable(self.ctx.BLEND)
            # Set custom blend function or revert to default
            if blend_function is not None:
                self.ctx.blend_func = blend_function
            else:
                self.ctx.blend_func = self.ctx.BLEND_DEFAULT
        else:
            self.ctx.disable(self.ctx.BLEND)

        # Workarounds for Optional[TextureAtlas] + slow . lookup speed
        atlas: DefaultTextureAtlas = self.atlas  # type: ignore
        atlas_texture: Texture2D = atlas.texture

        # Set custom filter or reset to default
        if filter:
            if hasattr(
                filter,
                "__len__",
            ):  # assume it's a collection
                if len(cast(Sized, filter)) != 2:
                    raise ValueError("Can't use sequence of length != 2")
                atlas_texture.filter = tuple(filter)  # type: ignore
            else:  # assume it's an int
                atlas_texture.filter = cast(OpenGlFilter, (filter, filter))
        else:
            # Handle the pixelated shortcut if filter is not set
            if pixelated:
                atlas_texture.filter = self.ctx.NEAREST, self.ctx.NEAREST
            else:
                atlas_texture.filter = self.DEFAULT_TEXTURE_FILTER

        self.program["spritelist_color"] = self._color

        # Control center pixel interpolation:
        # 0.0 = raw interpolation using texture corners
        # 1.0 = center pixel interpolation
        if self.ctx.NEAREST in atlas_texture.filter:
            self.program.set_uniform_safe("uv_offset_bias", 0.0)
        else:
            self.program.set_uniform_safe("uv_offset_bias", 1.0)

        atlas_texture.use(0)
        atlas.use_uv_texture(1)
        if not self._geometry:
            raise ValueError("Attempting to render without '_geometry' field being set.")
        self._geometry.render(
            self.program,
            mode=self.ctx.POINTS,
            vertices=self._sprite_index_slots,
        )

        # Leave global states to default
        if self._blend:
            self.ctx.disable(self.ctx.BLEND)
            if blend_function is not None:
                self.ctx.blend_func = prev_blend_func

    def draw_hit_boxes(
        self, color: RGBOrA255 = (0, 0, 0, 255), line_thickness: float = 1.0
    ) -> None:
        import arcade

        converted_color = Color.from_iterable(color)
        points: list[Point2] = []

        # TODO: Make this faster in the future
        # NOTE: This will be easier when/if we change to triangles
        for sprite in self.sprite_list:
            adjusted_points = sprite.hit_box.get_adjusted_points()
            for i in range(len(adjusted_points) - 1):
                points.append(adjusted_points[i])
                points.append(adjusted_points[i + 1])
            points.append(adjusted_points[-1])
            points.append(adjusted_points[0])

        arcade.draw_lines(points, color=converted_color, line_width=line_thickness)

    def _normalize_index_buffer(self) -> None:
        """
        Removes unused slots in the index buffer.
        The other buffers don't need this because they re-use slots.
        New sprites on the other hand always needs to be added
        to the end of the index buffer to preserve order
        """
        # NOTE: Currently we keep the index buffer normalized
        #       but we can increase the performance in the future
        #       delaying normalization.
        # Need counter for how many slots are used in index buffer.
        # 1) Sort the deleted indices (descending) and pop() them in a loop
        # 2) Create a new array.array and manually copy every
        #    item in the list except the deleted index slots
        # 3) Use a transform (gpu) to trim the index buffer and
        #    read this buffer back into a new array using array.from_bytes
        # NOTE: Right now the index buffer is always normalized
        pass

    def _grow_sprite_buffers(self) -> None:
        """Double the internal buffer sizes"""
        # Resize sprite buffers if needed
        if self._sprite_buffer_slots <= self._buf_capacity:
            return

        # Double the capacity
        extend_by = self._buf_capacity
        self._buf_capacity = self._buf_capacity * 2

        # Extend the buffers so we don't lose the old data
        self._sprite_pos_data.extend([0] * extend_by * 3)
        self._sprite_size_data.extend([0] * extend_by * 2)
        self._sprite_angle_data.extend([0] * extend_by)
        self._sprite_color_data.extend([0] * extend_by * 4)
        self._sprite_texture_data.extend([0] * extend_by)

        if self._initialized:
            # Proper initialization implies these buffers are allocated
            self._sprite_pos_buf.orphan(double=True)  # type: ignore
            self._sprite_size_buf.orphan(double=True)  # type: ignore
            self._sprite_angle_buf.orphan(double=True)  # type: ignore
            self._sprite_color_buf.orphan(double=True)  # type: ignore
            self._sprite_texture_buf.orphan(double=True)  # type: ignore

        self._sprite_pos_changed = True
        self._sprite_size_changed = True
        self._sprite_angle_changed = True
        self._sprite_color_changed = True
        self._sprite_texture_changed = True

    def _grow_index_buffer(self) -> None:
        # Extend the index buffer capacity if needed
        if self._sprite_index_slots <= self._idx_capacity:
            return

        extend_by = self._idx_capacity
        self._idx_capacity = self._idx_capacity * 2

        self._sprite_index_data.extend([0] * extend_by)
        if self._initialized and self._sprite_index_buf:
            self._sprite_index_buf.orphan(size=self._idx_capacity * 4)

        self._sprite_index_changed = True

    def _update_all(self, sprite: SpriteType) -> None:
        """
        Update all sprite data. This is faster when adding and moving sprites.
        This duplicate code, but reduces call overhead, dict lookups etc.

        Args:
            sprite: Sprite to update.
        """
        slot = self.sprite_slot[sprite]
        # position
        self._sprite_pos_data[slot * 3] = sprite._position[0]
        self._sprite_pos_data[slot * 3 + 1] = sprite._position[1]
        self._sprite_pos_data[slot * 3 + 2] = sprite._depth
        self._sprite_pos_changed = True
        # size
        self._sprite_size_data[slot * 2] = sprite._width
        self._sprite_size_data[slot * 2 + 1] = sprite._height
        self._sprite_size_changed = True
        # angle
        self._sprite_angle_data[slot] = sprite._angle
        self._sprite_angle_changed = True
        # color
        self._sprite_color_data[slot * 4] = sprite._color[0]
        self._sprite_color_data[slot * 4 + 1] = sprite._color[1]
        self._sprite_color_data[slot * 4 + 2] = sprite._color[2]
        self._sprite_color_data[slot * 4 + 3] = sprite._color[3] * sprite._visible
        self._sprite_color_changed = True

        # Don't deal with textures if spritelist is not initialized.
        # This can often mean we don't have a context/window yet.
        if not self._initialized:
            return

        if not sprite._texture:
            return

        # Ugly syntax makes type checking pass without perf hit from cast
        tex_slot: int = self._atlas.add(sprite._texture)[0]  # type: ignore
        slot = self.sprite_slot[sprite]

        self._sprite_texture_data[slot] = tex_slot
        self._sprite_texture_changed = True

    def _update_texture(self, sprite: SpriteType) -> None:
        """
        Make sure we update the texture for this sprite for the next batch drawing.

        Args:
            sprite: Sprite to update.
        """
        # We cannot interact with texture atlases unless the context
        # is created. We defer all texture initialization for later
        if not self._initialized:
            return

        if not sprite._texture:
            return
        atlas = self._atlas
        # Ugly spacing makes type checking work with specificity
        tex_slot: int = atlas.add(sprite._texture)[0]  # type: ignore
        slot = self.sprite_slot[sprite]

        self._sprite_texture_data[slot] = tex_slot
        self._sprite_texture_changed = True

        # Update size in cas the sprite was initialized without size
        # NOTE: There should be a better way to do this
        self._sprite_size_data[slot * 2] = sprite._width
        self._sprite_size_data[slot * 2 + 1] = sprite._height
        self._sprite_size_changed = True

    def _update_position(self, sprite: SpriteType) -> None:
        """
        Called when setting initial position of a sprite when
        added or inserted into the SpriteList.

        ``update_location`` should be called to move them
        once the sprites are in the list.

        Args:
            sprite: Sprite to update.
        """
        slot = self.sprite_slot[sprite]
        self._sprite_pos_data[slot * 3] = sprite._position[0]
        self._sprite_pos_data[slot * 3 + 1] = sprite._position[1]
        self._sprite_pos_changed = True

    def _update_position_x(self, sprite: SpriteType) -> None:
        """
        Called when setting initial position of a sprite when
        added or inserted into the SpriteList.

        ``update_location`` should be called to move them
        once the sprites are in the list.

        Args:
            sprite: Sprite to update.
        """
        slot = self.sprite_slot[sprite]
        self._sprite_pos_data[slot * 3] = sprite._position[0]
        self._sprite_pos_changed = True

    def _update_position_y(self, sprite: SpriteType) -> None:
        """
        Called when setting initial position of a sprite when
        added or inserted into the SpriteList.

        ``update_location`` should be called to move them
        once the sprites are in the list.

        Args:
            sprite: Sprite to update.
        """
        slot = self.sprite_slot[sprite]
        self._sprite_pos_data[slot * 3 + 1] = sprite._position[1]
        self._sprite_pos_changed = True

    def _update_depth(self, sprite: SpriteType) -> None:
        """
        Called by the Sprite class to update the depth of the specified sprite.
        Necessary for batch drawing of items.

        Args:
            sprite: Sprite to update.
        """
        slot = self.sprite_slot[sprite]
        self._sprite_pos_data[slot * 3 + 2] = sprite._depth
        self._sprite_pos_changed = True

    def _update_color(self, sprite: SpriteType) -> None:
        """
        Called by the Sprite class to update position, angle, size and color
        of the specified sprite.
        Necessary for batch drawing of items.

        Args:
            sprite: Sprite to update.
        """
        slot = self.sprite_slot[sprite]
        self._sprite_color_data[slot * 4] = int(sprite._color[0])
        self._sprite_color_data[slot * 4 + 1] = int(sprite._color[1])
        self._sprite_color_data[slot * 4 + 2] = int(sprite._color[2])
        self._sprite_color_data[slot * 4 + 3] = int(sprite._color[3] * sprite._visible)
        self._sprite_color_changed = True

    def _update_size(self, sprite: SpriteType) -> None:
        """
        Called by the Sprite class to update the size/scale in this sprite.
        Necessary for batch drawing of items.

        Args:
            sprite: Sprite to update.
        """
        slot = self.sprite_slot[sprite]
        self._sprite_size_data[slot * 2] = sprite._width
        self._sprite_size_data[slot * 2 + 1] = sprite._height
        self._sprite_size_changed = True

    def _update_width(self, sprite: SpriteType) -> None:
        """
        Called by the Sprite class to update the size/scale in this sprite.
        Necessary for batch drawing of items.

        Args:
            sprite: Sprite to update.
        """
        slot = self.sprite_slot[sprite]
        self._sprite_size_data[slot * 2] = sprite._width
        self._sprite_size_changed = True

    def _update_height(self, sprite: SpriteType) -> None:
        """
        Called by the Sprite class to update the size/scale in this sprite.
        Necessary for batch drawing of items.

        Args:
            sprite: Sprite to update.
        """
        slot = self.sprite_slot[sprite]
        self._sprite_size_data[slot * 2 + 1] = sprite._height
        self._sprite_size_changed = True

    def _update_angle(self, sprite: SpriteType) -> None:
        """
        Called by the Sprite class to update the angle in this sprite.
        Necessary for batch drawing of items.

        Args:
            sprite: Sprite to update.
        """
        slot = self.sprite_slot[sprite]
        self._sprite_angle_data[slot] = sprite._angle
        self._sprite_angle_changed = True



"""
Work with a mini-map

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.minimap_texture
"""

import random
from uuid import uuid4

import arcade

SPRITE_SCALING = 0.5

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
ASPECT_RATIO = WINDOW_WIDTH / WINDOW_HEIGHT
WINDOW_TITLE = "Minimap Example"

# How many pixels to keep as a minimum margin between the character
# and the edge of the screen.
VIEWPORT_MARGIN = 220

# How fast the camera pans to the player. 1.0 is instant.
CAMERA_SPEED = 0.1

# How fast the character moves
PLAYER_MOVEMENT_SPEED = 7

MINIMAP_BACKGROUND_COLOR = arcade.color.ALMOND
MINIMAP_WIDTH = 256
MINIMAP_HEIGHT = 256
MAP_WIDTH = 2048
MAP_HEIGHT = 2048


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        """
        Initializer
        """
        super().__init__()

        # Sprite lists
        self.player_list = None
        self.wall_list = None

        # Mini-map related
        # List of all our minimaps (there's just one)
        self.minimap_sprite_list = None
        # Texture and associated sprite to render our minimap to
        self.minimap_texture = None
        self.minimap_sprite = None

        # Set up the player
        self.player_sprite = None

        self.physics_engine = None

        # Camera for sprites, and one for our GUI
        self.camera_sprites = arcade.Camera2D()
        self.camera_gui = arcade.Camera2D()

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.player_list = arcade.SpriteList()
        self.wall_list = arcade.SpriteList()

        # Set up the player
        self.player_sprite = arcade.Sprite(
            ":resources:images/animated_characters/female_person/"
            "femalePerson_idle.png",
            scale=0.4,
        )
        self.player_sprite.center_x = 256
        self.player_sprite.center_y = 512
        self.player_list.append(self.player_sprite)

        # -- Set up several columns of walls
        for x in range(0, MAP_WIDTH, 210):
            for y in range(0, MAP_HEIGHT, 64):
                # Randomly skip a box so the player can find a way through
                if random.randrange(5) > 0:
                    wall = arcade.Sprite(
                        ":resources:images/tiles/grassCenter.png",
                        scale=SPRITE_SCALING,
                    )
                    wall.center_x = x
                    wall.center_y = y
                    self.wall_list.append(wall)

        self.physics_engine = arcade.PhysicsEngineSimple(self.player_sprite, self.wall_list)

        # Set the background color
        self.background_color = arcade.color.AMAZON

        # Construct the minimap
        size = (MINIMAP_WIDTH, MINIMAP_HEIGHT)
        self.minimap_texture = arcade.Texture.create_empty(str(uuid4()), size)
        self.minimap_sprite = arcade.Sprite(
            self.minimap_texture,
            center_x=MINIMAP_WIDTH / 2,
            center_y=self.height - MINIMAP_HEIGHT / 2,
        )

        self.minimap_sprite_list = arcade.SpriteList()
        self.minimap_sprite_list.append(self.minimap_sprite)

    def update_minimap(self):
        proj = 0, MAP_WIDTH, 0, MAP_HEIGHT
        atlas = self.minimap_sprite_list.atlas
        with atlas.render_into(self.minimap_texture, projection=proj) as fbo:
            fbo.clear(color=MINIMAP_BACKGROUND_COLOR)
            self.wall_list.draw()
            arcade.draw_sprite(self.player_sprite)

    def on_draw(self):
        """
        Render the screen.
        """

        # This command has to happen before we start drawing
        self.clear()

        # Select the camera we'll use to draw all our sprites
        with self.camera_sprites.activate():
            # Draw all the sprites.
            self.wall_list.draw()
            self.player_list.draw()

        # Select the (unscrolled) camera for our GUI
        with self.camera_gui.activate():
            # Update the minimap
            self.update_minimap()

            # Draw the minimap
            self.minimap_sprite_list.draw()

            # Draw the GUI
            arcade.draw_rect_filled(
                arcade.rect.XYWH(self.width // 2, 20, self.width, 40),
                color=arcade.color.ALMOND,
            )

            text = (
                f"Scroll value: "
                f"{self.camera_sprites.position[0]:4.1f}, "
                f"{self.camera_sprites.position[1]:4.1f}"
            )
            arcade.draw_text(text, 10, 10, arcade.color.BLACK_BEAN, 20)

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """

        if key == arcade.key.UP:
            self.player_sprite.change_y = PLAYER_MOVEMENT_SPEED
        elif key == arcade.key.DOWN:
            self.player_sprite.change_y = -PLAYER_MOVEMENT_SPEED
        elif key == arcade.key.LEFT:
            self.player_sprite.change_x = -PLAYER_MOVEMENT_SPEED
        elif key == arcade.key.RIGHT:
            self.player_sprite.change_x = PLAYER_MOVEMENT_SPEED

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key. """

        if key == arcade.key.UP or key == arcade.key.DOWN:
            self.player_sprite.change_y = 0
        elif key == arcade.key.LEFT or key == arcade.key.RIGHT:
            self.player_sprite.change_x = 0

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.physics_engine.update()

        # Scroll the screen to the player
        self.scroll_to_player()

    def scroll_to_player(self):
        """
        Scroll the window to the player.
        """

        # Scroll to the proper location
        position = (self.player_sprite.center_x, self.player_sprite.center_y)
        self.camera_sprites.position = arcade.math.lerp_2d(
            self.camera_sprites.position,
            position,
            CAMERA_SPEED,
        )

    def on_resize(self, width: int, height: int):
        """
        Resize window
        Handle the user grabbing the edge and resizing the window.
        """
        super().on_resize(width, height)
        self.camera_sprites.match_window(aspect=ASPECT_RATIO, projection=False)
        self.camera_gui.match_window(aspect=ASPECT_RATIO, projection=False)


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE, resizable=True)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Example "Arcade" library code.

This example shows the drawing primitives and how they are used.
It does not assume the programmer knows how to define functions or classes
yet.

API documentation for the draw commands can be found here:
https://api.arcade.academy/en/latest/quick_index.html

A video explaining this example can be found here:
https://vimeo.com/167158158

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.drawing_primitives
"""

# Import the Arcade library. If this fails, then try following the instructions
# for how to install Arcade:
# https://api.arcade.academy/en/latest/install/index.html
import arcade

# Open the window. Set the window title and dimensions (width and height)
arcade.open_window(600, 600, "Drawing Primitives Example", resizable=True)

# Set the background color to white
# For a list of named colors see
# https://api.arcade.academy/en/latest/arcade.color.html
# Colors can also be specified in (red, green, blue) format and
# (red, green, blue, alpha) format.
arcade.set_background_color(arcade.color.WHITE)

# Start the render process. This must be done before any drawing commands.
arcade.start_render()

# Draw a grid
# Draw vertical lines every 120 pixels
for x in range(0, 601, 120):
    arcade.draw_line(x, 0, x, 600, arcade.color.BLACK, 2)

# Draw horizontal lines every 200 pixels
for y in range(0, 601, 200):
    arcade.draw_line(0, y, 800, y, arcade.color.BLACK, 2)

# Draw a point
arcade.draw_text("draw_point", 3, 405, arcade.color.BLACK, 12)
arcade.draw_point(60, 495, arcade.color.RED, 10)

# Draw a set of points
arcade.draw_text("draw_points", 123, 405, arcade.color.BLACK, 12)
point_list = ((165, 495),
              (165, 480),
              (165, 465),
              (195, 495),
              (195, 480),
              (195, 465))
arcade.draw_points(point_list, arcade.color.ZAFFRE, 10)

# Draw a line
arcade.draw_text("draw_line", 243, 405, arcade.color.BLACK, 12)
arcade.draw_line(270, 495, 300, 450, arcade.color.WOOD_BROWN, 3)

# Draw a set of lines
arcade.draw_text("draw_lines", 363, 405, arcade.color.BLACK, 12)
point_list = ((390, 450),
              (450, 450),
              (390, 480),
              (450, 480),
              (390, 510),
              (450, 510)
              )
arcade.draw_lines(point_list, arcade.color.BLUE, 3)

# Draw a line strip
arcade.draw_text("draw_line_strip", 483, 405, arcade.color.BLACK, 12)
point_list = ((510, 450),
              (570, 450),
              (510, 480),
              (570, 480),
              (510, 510),
              (570, 510)
              )
arcade.draw_line_strip(point_list, arcade.color.TROPICAL_RAIN_FOREST, 3)

# Draw a polygon
arcade.draw_text("draw_polygon_outline", 3, 207, arcade.color.BLACK, 9)
point_list = ((30, 240),
              (45, 240),
              (60, 255),
              (60, 285),
              (45, 300),
              (30, 300))
arcade.draw_polygon_outline(point_list, arcade.color.SPANISH_VIOLET, 3)

# Draw a filled in polygon
arcade.draw_text("draw_polygon_filled", 123, 207, arcade.color.BLACK, 9)
point_list = ((150, 240),
              (165, 240),
              (180, 255),
              (180, 285),
              (165, 300),
              (150, 300))
arcade.draw_polygon_filled(point_list, arcade.color.SPANISH_VIOLET)

# Draw an outline of a circle
arcade.draw_text("draw_circle_outline", 243, 207, arcade.color.BLACK, 10)
arcade.draw_circle_outline(300, 285, 18, arcade.color.WISTERIA, 3)

# Draw a filled in circle
arcade.draw_text("draw_circle_filled", 363, 207, arcade.color.BLACK, 10)
arcade.draw_circle_filled(420, 285, 18, arcade.color.GREEN)

# Draw an ellipse outline, and another one rotated
arcade.draw_text("draw_ellipse_outline", 483, 207, arcade.color.BLACK, 10)
arcade.draw_ellipse_outline(540, 273, 15, 36, arcade.color.AMBER, 3)
arcade.draw_ellipse_outline(540, 336, 15, 36,
                            arcade.color.BLACK_BEAN, 3, 45)

# Draw a filled ellipse, and another one rotated
arcade.draw_text("draw_ellipse_filled", 3, 3, arcade.color.BLACK, 10)
arcade.draw_ellipse_filled(60, 81, 15, 36, arcade.color.AMBER)
arcade.draw_ellipse_filled(60, 144, 15, 36,
                           arcade.color.BLACK_BEAN, 45)

# Draw an arc, and another one rotated
arcade.draw_text("draw_arc/filled_arc", 123, 3, arcade.color.BLACK, 10)
arcade.draw_arc_outline(150, 81, 15, 36,
                        arcade.color.BRIGHT_MAROON, 90, 360)
arcade.draw_arc_filled(150, 144, 15, 36,
                       arcade.color.BOTTLE_GREEN, 90, 360, 45)

# Draw an rectangle outline
arcade.draw_text("draw_rect", 243, 3, arcade.color.BLACK, 10)
arcade.draw_rect_outline(arcade.rect.XYWH(295, 100, 45, 65),
                         arcade.color.BRITISH_RACING_GREEN)
arcade.draw_rect_outline(arcade.rect.XYWH(295, 160, 20, 45),
                         arcade.color.BRITISH_RACING_GREEN, 3, 45)

# Draw a filled in rectangle
arcade.draw_text("draw_filled_rect", 363, 3, arcade.color.BLACK, 10)
arcade.draw_rect_filled(arcade.rect.XYWH(420, 100, 45, 65), arcade.color.BLUSH)
arcade.draw_rect_filled(arcade.rect.XYWH(420, 160, 20, 40), arcade.color.BLUSH, 45)

# Load and draw an image to the screen
# Image from kenney.nl asset pack #1
arcade.draw_text("draw_bitmap", 483, 3, arcade.color.BLACK, 12)
texture = arcade.load_texture(":resources:images/space_shooter/playerShip1_orange.png")
scale = .6
arcade.draw_texture_rect(
    texture,
    arcade.XYWH(540, 120, texture.width, texture.height).scale(scale)
)
arcade.draw_texture_rect(
    texture,
    arcade.XYWH(540, 60, texture.width, texture.height).scale(scale), angle=45,
)

# Finish the render.
# Nothing will be drawn without this.
# Must happen after all draw commands
arcade.finish_render()

# Keep the window up until someone closes it.
arcade.run()



"""
THIS IS WORK IN PROGRESS. DO NOT USE.
"""

from typing import TYPE_CHECKING

from .base import TextureAtlasBase

if TYPE_CHECKING:
    from arcade import ArcadeContext


class BindlessTextureAtlas(TextureAtlasBase):
    """
    A texture atlas using bindless textures. This is an extension to OpenGL 4.3+
    and requires the ARB_bindless_texture extension or possibly vendor specific ones.

    Bindless textures are simply 64 bit integer names/handles for a texture that
    can be casted to a sampler in a shader meaning that you can have an unlimited
    amount of samplers in the same draw call. Likely a shader storage buffer is
    needed to store the handles.

    The max capacity is only limited by memory. Discrete GPUs support this
    feature and some iGPUs from 2019 and later. AMD + Intel Xe if
    latest drivers are installed.
    """

    def __init__(self, ctx: ArcadeContext):
        super().__init__(ctx)



"""An experimental depth-of-field example.

It uses the depth attribute of along with blurring and shaders to
roughly approximate depth-based blur effects. The focus bounces
back forth automatically between a maximum and minimum depth value
based on time. Change the speed and focus via either the constants
at the top of the file or the arguments passed to it at the bottom of
the file.

This example works by doing the following for each frame:

1. Render a depth value for pixel into a buffer
2. Render a gaussian blurred version of the scene
3. For each pixel, use the current depth value to lerp between the
   blurred and un-blurred versions of the scene.

This is more expensive than rendering the scene directly, but it's
both easier and more performant than more accurate blur approaches.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.depth_of_field
"""

from contextlib import contextmanager
from math import cos, pi
from random import randint, uniform
from textwrap import dedent
from typing import cast

from pyglet.graphics import Batch

import arcade
from arcade import get_window, SpriteList, SpriteSolidColor, Text, Window, View
from arcade.camera.data_types import DEFAULT_NEAR_ORTHO, DEFAULT_FAR
from arcade.color import RED
from arcade.experimental.postprocessing import GaussianBlur
from arcade.gl import NEAREST, Program, Texture2D, geometry
from arcade.types import RGBA255, Color

WINDOW_TITLE = "Depth of Field Example"

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
BACKGROUND_GRAY = Color(155, 155, 155, 255)


class DepthOfField:
    """A depth-of-field effect we can use as a render context manager.

    Args:
        size:
            The size of the buffers.
        clear_color:
            The color which will be used as the background.
    """

    def __init__(
        self,
        size: tuple[int, int] | None = None,
        clear_color: RGBA255 = BACKGROUND_GRAY
    ):
        self._geo = geometry.quad_2d_fs()
        self._win: Window = get_window()

        size = cast(tuple[int, int], size or self._win.size)
        self._clear_color: Color = Color.from_iterable(clear_color)

        self.stale = True

        # Set up our depth buffer to hold per-pixel depth
        self._render_target = self._win.ctx.framebuffer(
            color_attachments=[
                self._win.ctx.texture(
                    size,
                    components=4,
                    filter=(NEAREST, NEAREST),
                    wrap_x=self._win.ctx.REPEAT,
                    wrap_y=self._win.ctx.REPEAT,
                ),
            ],
            depth_attachment=self._win.ctx.depth_texture(size),
        )

        # Set up everything we need to perform blur and store results.
        # This includes the blur effect, a framebuffer, and an instance
        # variable to store the returned texture holding blur results.
        self._blur_process = GaussianBlur(size, kernel_size=10, sigma=2.0, multiplier=2.0, step=4)
        self._blur_target = self._win.ctx.framebuffer(
            color_attachments=[
                self._win.ctx.texture(
                    size,
                    components=4,
                    filter=(NEAREST, NEAREST),
                    wrap_x=self._win.ctx.REPEAT,
                    wrap_y=self._win.ctx.REPEAT,
                )
            ]
        )
        self._blurred: Texture2D | None = None

        # To keep this example in one file, we use strings for our
        # our shaders. You may want to use pathlib.Path.read_text in
        # your own code instead.
        self._render_program = self._win.ctx.program(
            vertex_shader=dedent(
                """#version 330

                in vec2 in_vert;
                in vec2 in_uv;

                out vec2 out_uv;

                void main(){
                   gl_Position = vec4(in_vert, 0.0, 1.0);
                   out_uv = in_uv;
                }"""
            ),
            fragment_shader=dedent(
                """#version 330

                uniform sampler2D texture_0;
                uniform sampler2D texture_1;
                uniform sampler2D depth_0;

                uniform float focus_depth;

                in vec2 out_uv;

                out vec4 frag_colour;

                void main() {
                   float depth_val = texture(depth_0, out_uv).x;
                   float depth_adjusted = min(1.0, 2.0 * abs(depth_val - focus_depth));
                   vec4 crisp_tex = texture(texture_0, out_uv);
                   vec3 blur_tex = texture(texture_1, out_uv).rgb;
                   frag_colour = mix(crisp_tex, vec4(blur_tex, crisp_tex.a), depth_adjusted);
                   //if (depth_adjusted < 0.1){frag_colour = vec4(1.0, 0.0, 0.0, 1.0);}
                }"""
            ),
        )

        # Set the buffers the shader program will use
        self._render_program["texture_0"] = 0
        self._render_program["texture_1"] = 1
        self._render_program["depth_0"] = 2

    @property
    def render_program(self) -> Program:
        """The compiled shader for this effect."""
        return self._render_program

    @contextmanager
    def draw_into(self):
        self.stale = True
        previous_fbo = self._win.ctx.active_framebuffer
        try:
            self._win.ctx.enable(self._win.ctx.DEPTH_TEST)
            self._render_target.clear(color=self._clear_color)
            self._render_target.use()
            yield self._render_target
        finally:
            self._win.ctx.disable(self._win.ctx.DEPTH_TEST)
            previous_fbo.use()

    def process(self):
        self._blurred = self._blur_process.render(self._render_target.color_attachments[0])
        self._win.use()

        self.stale = False

    def render(self):
        if self.stale:
            self.process()

        self._render_target.color_attachments[0].use(0)
        self._blurred.use(1)
        self._render_target.depth_attachment.use(2)
        self._geo.render(self._render_program)


class GameView(View):
    """Window subclass to hold sprites and rendering helpers.

    To keep the code simpler, this example uses a default camera. That means
    that any sprite outside Arcade's default camera near and far render cutoffs
    (``-100.0`` to ``100.0``) will not be drawn.

    Args:
        text_color:
            The color of the focus indicator.
        focus_range:
            The range the focus value will oscillate between.
        focus_change_speed:
            How fast the focus bounces back and forth
            between the ``-focus_range`` and ``focus_range``.
        min_sprite_depth:
            The minimum sprite depth we'll generate sprites between
         max_sprite_depth:
            The maximum sprite depth we'll generate sprites between.
    """

    def __init__(
        self,
        text_color: RGBA255 = RED,
        focus_range: float = 16.0,
        focus_change_speed: float = 0.1,
        min_sprite_depth: float = DEFAULT_NEAR_ORTHO,
        max_sprite_depth: float = DEFAULT_FAR
    ):
        super().__init__()
        self.sprites: SpriteList = SpriteList()
        self._batch = Batch()
        self.focus_range: float = focus_range
        self.focus_change_speed: float = focus_change_speed
        self.indicator_label = Text(
            f"Focus depth: {0:.3f} / {focus_range}",
            self.width / 2,
            self.height / 2,
            text_color,
            align="center",
            anchor_x="center",
            batch=self._batch,
        )

        # Randomize sprite depth, size, and angle, but set color from depth.
        for _ in range(100):
            depth = uniform(min_sprite_depth, max_sprite_depth)
            color = Color.from_gray(int(255 * (depth + 100) / 200))
            s = SpriteSolidColor(
                randint(100, 200),
                randint(100, 200),
                uniform(20, self.width - 20),
                uniform(20, self.height - 20),
                color,
                uniform(0, 360),
            )
            s.depth = depth
            self.sprites.append(s)

        self.dof = DepthOfField()

    def on_update(self, delta_time: float):
        time = self.window.time
        raw_focus = self.focus_range * (cos(pi * self.focus_change_speed * time) * 0.5 + 0.5)
        self.dof.render_program["focus_depth"] = raw_focus / self.focus_range
        self.indicator_label.value = f"Focus depth: {raw_focus:.3f} / {self.focus_range}"

    def on_draw(self):
        self.clear()

        # Render the depth-of-field layer's frame buffer
        with self.dof.draw_into():
            self.sprites.draw(pixelated=True)

        # Draw the blurred frame buffer and then the focus display
        window = self.window
        window.use()
        self.dof.render()
        self._batch.draw()


def main():
    # Create a Window to show the view defined above.
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the view
    app = GameView()

    # Show GameView on screen
    window.show_view(app)

    # Start the arcade game loop
    window.run()


if __name__ == "__main__":
    main()



"""
Shows how we can use shaders using existing spritelist data.

This example builds on previous examples visualizing selection
of nearby sprites taking line of sight into consideration.

This version on the other hand does not visualize. We want
return useful data from the shader to resolve the actual
selected sprite objects so we can use it in our game logic.

For this to happen we need to change our shader into a
transform shader. These shaders do not draw to the screen.
They write data into a buffer we can read back on the
python side. This simply involves removing the fragment
shader and instead defining some values to output.
We also need to do a transform() instead of render()
and supply a buffer for our results. In addition we
use a Query to count how many results the shader gave us.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.spritelist_interaction_visualize_dist_los_trans
"""

import random
import struct
import arcade

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
NUM_COINS = 500
NUM_WALLS = 100
INTERACTION_RADIUS = 200


class SpriteListInteraction(arcade.Window):

    def __init__(self):
        super().__init__(WINDOW_WIDTH, WINDOW_HEIGHT, "SpriteList Interaction - LoS")

        # Player
        self.player = arcade.Sprite(
            ":resources:images/animated_characters/female_person/femalePerson_idle.png",
            scale=0.25,
        )

        # Wall sprites we are checking collision against
        self.walls = arcade.SpriteList()
        for _ in range(NUM_WALLS):
            self.walls.append(
                arcade.Sprite(
                    ":resources:images/tiles/boxCrate_double.png",
                    center_x=random.randint(0, WINDOW_WIDTH),
                    center_y=random.randint(0, WINDOW_HEIGHT),
                    scale=0.25,
                )
            )

        # Generate some random coins.
        # We make sure they are not placed inside a wall.
        # We give each coin 1 chance to spawn at the right position.
        self.coins = arcade.SpriteList()
        for _ in range(NUM_COINS):
            coin = arcade.Sprite(
                ":resources:images/items/coinGold.png",
                center_x=random.randint(0, WINDOW_WIDTH),
                center_y=random.randint(0, WINDOW_HEIGHT),
                scale=0.25,
            )
            if arcade.check_for_collision_with_list(coin, self.walls):
                continue

            self.coins.append(coin)

        # This program finds sprites within a certain distance
        # not blocked by line of sight. We use a texture containing
        # the walls to trace a ray between the player and coin.
        # If any non-zero pixels are detected we skip the sprite.
        self.program_select_sprites = self.ctx.program(
            vertex_shader="""
            #version 330

            // Sprite positions from SpriteList
            in vec3 in_pos;

            // Output to geometry shader
            out vec3 v_position;

            void main() {
                // This shader just forwards info to geo shader
                v_position = in_pos;
            }
            """,
            geometry_shader="""
            #version 330

            // The position we measure distance from
            uniform vec2 origin;
            // The maximum distance
            uniform float maxDistance;
            // Sampler for reading wall data
            uniform sampler2D walls;

            // These configure the geometry shader to process a points
            // and also emit single points of data
            // in the spritelist.
            layout (points) in;
            layout (points, max_vertices = 1) out;

            // The position input from vertex shader.
            // It's an array because geo shader can take more than one input
            in vec3 v_position[];
            // This shader outputs all the sprite indices it selected.
            // NOTE: We use floats here for compatibility since integers are always working
            out float spriteIndex;

            // Helper function converting screen coordinates to texture coordinates.
            // Texture coordinates are normalized (0.0 -> 1.0) were 0,0 is in the
            vec2 screen2texcoord(vec2 pos) {
                return vec2(pos / vec2(textureSize(walls, 0).xy));
            }

            void main() {
                // ONLY emit a line between the sprite and origin when within the distance
                if (distance(v_position[0].xy, origin) > maxDistance) return;

                // Read samples from the wall texture in a line looking for obstacles
                // We simply make a vector between the original and the sprite location
                // and trace pixels in this path with a reasonable step.
                int numSteps = int(maxDistance / 2.0);
                vec2 dir = v_position[0].xy - origin;
                for (int i = 0; i < numSteps; i++) {
                    // Read pixels along the vector
                    vec2 pos = origin + dir * (float(i) / float(numSteps));
                    vec4 color = texture(walls, screen2texcoord(pos));
                    // If we find non-zero pixel data we have obstacles in our path!
                    if (color != vec4(0.0)) return;
                }
                // We simply return the primitive index.
                // This is a built in counter in geometry shaders
                // started at 0 incrementing by 1 for every invocation.
                // It should always match the spritelist index.
                spriteIndex = float(gl_PrimitiveIDIn);
                EmitVertex();
                EndPrimitive();
            }
            """,
        )
        # Configure program with maximum distance
        self.program_select_sprites["maxDistance"] = INTERACTION_RADIUS

        # NOTE: Before we do this we need to ensure the internal sprite buffers are complete!
        self.coins.write_sprite_buffers_to_gpu()
        # We need a buffer that can capture the output from our transform shader.
        # We need to make room for len(coins) 32 bit floats (4 bytes each. size specified in bytes)
        self.result_buffer = self.ctx.buffer(reserve=len(self.coins) * 4)
        # We need a query to count how many sprites the shader gave us.
        # We do this by making a sampler that counts the number of primitives
        # (points in this instance) it emitted into the result buffer.
        self.query = self.ctx.query()

        # Lookup texture/framebuffer for walls so we can trace pixels in the shader.
        # It contains a texture attachment with the same size as the window.
        # We draw only the walls into this one as a line of sight lookup
        self.walls_fbo = self.ctx.framebuffer(
            color_attachments=[self.ctx.texture((WINDOW_WIDTH, WINDOW_HEIGHT))]
        )
        # Draw the walls into the framebuffer
        with self.walls_fbo.activate() as fbo:
            fbo.clear()
            self.walls.draw()

    def on_draw(self):
        self.clear()

        # As long as we have coins...
        if len(self.coins) > 0:
            # Ensure the internal sprite buffers are up to date
            self.coins.write_sprite_buffers_to_gpu()
            # Bind the wall texture to texture channel 0 so we can read it in the shader
            self.walls_fbo.color_attachments[0].use(0)
            with self.query:
                # We already have a geometry instance in the spritelist we can
                # use to run our shader/gpu program. It only requires that we
                # use correctly named input name(s). in_pos in this example
                # what will automatically map in the position buffer to the vertex shader.
                self.coins.geometry.transform(
                    self.program_select_sprites,
                    self.result_buffer,
                    vertices=len(self.coins),
                )

            # Store the number of primitives/sprites found
            num_sprites_found = self.query.primitives_generated
            if num_sprites_found > 0:
                print(f"We found {num_sprites_found} sprite(s)")
                # Transfer the the data from the vram into python and decode the
                # values into python objects. We read num_sprites_found float
                # values from the result buffer and convert those into python floats
                sprite_indices = struct.unpack(
                    f"{num_sprites_found}f",
                    self.result_buffer.read(size=num_sprites_found * 4),
                )
                print("Indices found:", sprite_indices)
                print(

                        f"max(sprite_indices) = {max(sprite_indices)} | "
                        f"len(self.coins) = {len(self.coins)} | "
                        f"sprite_indices = {len(sprite_indices)}"

                )
                # Resolve the list of selected sprites and remove them
                sprites = [self.coins[int(i)] for i in sprite_indices]
                for coin in sprites:
                    coin.remove_from_sprite_lists()

        self.walls.draw()
        self.coins.draw()
        arcade.draw_sprite(self.player)

        # Visualize the interaction radius
        arcade.draw_circle_filled(
            center_x=self.player.center_x,
            center_y=self.player.center_y,
            radius=INTERACTION_RADIUS,
            color=(255, 255, 255, 64)
        )

    def on_mouse_motion(self, x: float, y: float, dx: float, dy: float):
        # Move the sprite to mouse position
        self.player.position = x, y
        # Update the program with a new origin
        self.program_select_sprites["origin"] = x, y


SpriteListInteraction().run()



"""
Tetris

Tetris clone, with some ideas from silvasur's code:
https://gist.github.com/silvasur/565419/d9de6a84e7da000797ac681976442073045c74a4

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.tetris
"""
import arcade
from arcade.clock import GLOBAL_CLOCK
import random
import PIL

# Set how many rows and columns we will have
ROW_COUNT = 24
COLUMN_COUNT = 10

# This sets the WIDTH and HEIGHT of each grid location
WIDTH = 30
HEIGHT = 30

# This sets the margin between each cell
# and on the edges of the screen.
MARGIN = 5

# Do the math to figure out our screen dimensions
WINDOW_WIDTH = (WIDTH + MARGIN) * COLUMN_COUNT + MARGIN
WINDOW_HEIGHT = (HEIGHT + MARGIN) * ROW_COUNT + MARGIN
WINDOW_TITLE = "Tetris"

colors = [
    (0,   0,   0, 255),
    (255, 0,   0, 255),
    (0,   150, 0, 255),
    (0,   0,   255, 255),
    (255, 120, 0, 255),
    (255, 255, 0, 255),
    (180, 0,   255, 255),
    (0,   220, 220, 255)
]

# Define the shapes of the single parts
tetris_shapes = [
    [[1, 1, 1],
     [0, 1, 0]],

    [[0, 2, 2],
     [2, 2, 0]],

    [[3, 3, 0],
     [0, 3, 3]],

    [[4, 0, 0],
     [4, 4, 4]],

    [[0, 0, 5],
     [5, 5, 5]],

    [[6, 6, 6, 6]],

    [[7, 7],
     [7, 7]]
]


def create_textures():
    """ Create a list of images for sprites based on the global colors. """
    new_textures = []
    for color in colors:
        image = PIL.Image.new('RGBA', (WIDTH, HEIGHT), color)
        new_textures.append(arcade.Texture(image))
    return new_textures


texture_list = create_textures()


def rotate_counterclockwise(shape):
    """ Rotates a matrix clockwise """
    return [[shape[y][x] for y in range(len(shape))]
            for x in range(len(shape[0]) - 1, -1, -1)]


def check_collision(board, shape, offset):
    """
    See if the matrix stored in the shape will intersect anything
    on the board based on the offset. Offset is an (x, y) coordinate.
    """
    off_x, off_y = offset
    for cy, row in enumerate(shape):
        for cx, cell in enumerate(row):
            if cell and board[cy + off_y][cx + off_x]:
                return True
    return False


def remove_row(board, row):
    """ Remove a row from the board, add a blank row on top. """
    del board[row]
    return [[0 for _ in range(COLUMN_COUNT)]] + board


def join_matrixes(matrix_1, matrix_2, matrix_2_offset):
    """ Copy matrix 2 onto matrix 1 based on the passed in x, y offset coordinate """
    offset_x, offset_y = matrix_2_offset
    for cy, row in enumerate(matrix_2):
        for cx, val in enumerate(row):
            matrix_1[cy + offset_y - 1][cx + offset_x] += val
    return matrix_1


def new_board():
    """ Create a grid of 0's. Add 1's to the bottom for easier collision detection. """
    # Create the main board of 0's
    board = [[0 for _x in range(COLUMN_COUNT)] for _y in range(ROW_COUNT)]
    # Add a bottom border of 1's
    board += [[1 for _x in range(COLUMN_COUNT)]]
    return board


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        super().__init__()
        self.window.background_color = arcade.color.WHITE

        self.board = None
        self.start_frame = 0
        self.game_over = False
        self.paused = False
        self.board_sprite_list = None

        self.stone = None
        self.stone_x = 0
        self.stone_y = 0

        self.stones = []

    def new_stone(self):
        """
        Randomly grab a new stone from the bag of stones,
        if the bag is empty refill the bag and shuffle it.
        Then set the stone's location to the top.
        If we immediately collide, then game-over.
        """
        if not self.stones:
            self.stones = tetris_shapes[:]

        random.shuffle(self.stones)
        self.stone = self.stones.pop()
        self.stone_x = int(COLUMN_COUNT / 2 - len(self.stone[0]) / 2)
        self.stone_y = 0

        if check_collision(self.board, self.stone, (self.stone_x, self.stone_y)):
            self.game_over = True

    def setup(self):
        self.board = new_board()
        self.start_frame = GLOBAL_CLOCK.ticks

        self.board_sprite_list = arcade.SpriteList()
        for row in range(len(self.board)):
            for column in range(len(self.board[0])):
                sprite = arcade.Sprite(texture_list[0])
                sprite.textures = texture_list
                sprite.center_x = (MARGIN + WIDTH) * column + MARGIN + WIDTH // 2
                sprite.center_y = WINDOW_HEIGHT - (MARGIN + HEIGHT) * row + MARGIN + HEIGHT // 2

                self.board_sprite_list.append(sprite)

        self.new_stone()
        self.update_board()

    def drop(self):
        """
        Drop the stone down one place.
        Check for collision.
        If collided, then
          join matrixes
          Check for rows we can remove
          Update sprite list with stones
          Create a new stone
        """
        if not self.game_over and not self.paused:
            self.stone_y += 1
            if check_collision(self.board, self.stone, (self.stone_x, self.stone_y)):
                self.board = join_matrixes(self.board, self.stone, (self.stone_x, self.stone_y))
                while True:
                    for i, row in enumerate(self.board[:-1]):
                        if 0 not in row:
                            self.board = remove_row(self.board, i)
                            break
                    else:
                        break
                self.update_board()
                self.new_stone()

    def rotate_stone(self):
        """ Rotate the stone, check collision. """
        if not self.game_over and not self.paused:
            new_stone = rotate_counterclockwise(self.stone)
            if self.stone_x + len(new_stone[0]) >= COLUMN_COUNT:
                self.stone_x = COLUMN_COUNT - len(new_stone[0])
            if not check_collision(self.board, new_stone, (self.stone_x, self.stone_y)):
                self.stone = new_stone

    def on_update(self, delta_time):
        """ Update, drop stone if warranted """
        if GLOBAL_CLOCK.ticks_since(self.start_frame) % 10 == 0:
            self.drop()

    def move(self, delta_x):
        """ Move the stone back and forth based on delta x. """
        if not self.game_over and not self.paused:
            new_x = self.stone_x + delta_x
            if new_x < 0:
                new_x = 0
            if new_x > COLUMN_COUNT - len(self.stone[0]):
                new_x = COLUMN_COUNT - len(self.stone[0])
            if not check_collision(self.board, self.stone, (new_x, self.stone_y)):
                self.stone_x = new_x

    def on_key_press(self, key, modifiers):
        """
        Handle user key presses
        User goes left, move -1
        User goes right, move 1
        Rotate stone,
        or drop down
        """
        if key == arcade.key.LEFT:
            self.move(-1)
        elif key == arcade.key.RIGHT:
            self.move(1)
        elif key == arcade.key.UP:
            self.rotate_stone()
        elif key == arcade.key.DOWN:
            self.drop()

    def draw_grid(self, grid, offset_x, offset_y):
        """
        Draw the grid. Used to draw the falling stones. The board is drawn
        by the sprite list.
        """
        # Draw the grid
        for row in range(len(grid)):
            for column in range(len(grid[0])):
                # Figure out what color to draw the box
                if grid[row][column]:
                    color = colors[grid[row][column]]
                    # Do the math to figure out where the box is
                    x = (
                        (MARGIN + WIDTH) * (column + offset_x)
                        + MARGIN + WIDTH // 2
                    )
                    y = (
                        WINDOW_HEIGHT - (MARGIN + HEIGHT) * (row + offset_y)
                        + MARGIN + HEIGHT // 2
                    )

                    # Draw the box
                    arcade.draw_rect_filled(arcade.rect.XYWH(x, y, WIDTH, HEIGHT), color)

    def update_board(self):
        """
        Update the sprite list to reflect the contents of the 2d grid
        """
        for row in range(len(self.board)):
            for column in range(len(self.board[0])):
                v = self.board[row][column]
                i = row * COLUMN_COUNT + column
                self.board_sprite_list[i].set_texture(v)

    def on_draw(self):
        """ Render the screen. """

        # This command has to happen before we start drawing
        self.clear()
        self.board_sprite_list.draw()
        self.draw_grid(self.stone, self.stone_x, self.stone_y)


def main():
    """ Create the game window, setup, run """
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    game = GameView()
    game.setup()

    window.show_view(game)
    arcade.run()


if __name__ == "__main__":
    main()



"""
Sprite Collect Coins

A simple game demonstrating an easy way to create and use sprites.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the
command line with:
python -m arcade.examples.sprite_collect_coins
"""

import random
import arcade

# --- Constants ---
SPRITE_SCALING_PLAYER = 0.5
SPRITE_SCALING_COIN = 0.4
COIN_COUNT = 50

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprite Collect Coins Example"


class GameView(arcade.View):

    def __init__(self):
        """ Initializer """
        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.player_list = None
        self.coin_list = None

        # Create a variable to hold the player sprite
        self.player_sprite = None

        # Variables to hold the score and the Text object displaying it
        self.score = 0
        self.score_display = None

        # Hide the mouse cursor while it's over the window
        self.window.set_mouse_visible(False)

        self.background_color = arcade.color.AMAZON

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Create the sprite lists
        self.player_list = arcade.SpriteList()
        self.coin_list = arcade.SpriteList()

        # Reset the score and the score display
        self.score = 0
        self.score_display = arcade.Text(
            text="Score: 0", x=10, y=20,
            color=arcade.color.WHITE, font_size=14)

        # Set up the player
        # Character image from kenney.nl
        img = ":resources:images/animated_characters/female_person/femalePerson_idle.png"
        self.player_sprite = arcade.Sprite(img, scale=SPRITE_SCALING_PLAYER)
        self.player_sprite.position = 50, 50
        self.player_list.append(self.player_sprite)

        # Create the coins
        for i in range(COIN_COUNT):

            # Create the coin instance
            # Coin image from kenney.nl
            coin = arcade.Sprite(":resources:images/items/coinGold.png",
                                 scale=SPRITE_SCALING_COIN)

            # Position the coin
            coin.center_x = random.randrange(WINDOW_WIDTH)
            coin.center_y = random.randrange(WINDOW_HEIGHT)

            # Add the coin to the lists
            self.coin_list.append(coin)

    def on_draw(self):
        """ Draw everything """

        # Clear the screen to only show the background color
        self.clear()

        # Draw the sprites
        self.coin_list.draw()
        self.player_list.draw()

        # Draw the score Text object on the screen
        self.score_display.draw()

    def on_mouse_motion(self, x, y, dx, dy):
        """ Handle Mouse Motion """

        # Move the player sprite to place its center on the mouse x, y
        self.player_sprite.position = x, y

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Generate a list of all sprites that collided with the player.
        coins_hit_list = arcade.check_for_collision_with_list(self.player_sprite,
                                                              self.coin_list)

        # Keep track of the score from before collisions occur
        old_score = self.score

        # Loop through each colliding sprite, remove it, and add to the score.
        for coin in coins_hit_list:
            coin.remove_from_sprite_lists()
            self.score += 1

        # Update the score display if the score changed this tick
        if old_score != self.score:
            self.score_display.text = f"Score: {self.score}"


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Game of Life - Shader Version

This version uses shaders which draws to textures to
run Conway's Game of Life with an added twist: colors.

Press SPACE to reset the simulation with random data.

It uses two textures: One to keep the old state and
a second to draw the new state into. These textures are
flipped around after every update.

You can configure the cell and window size by changing
the constants at the top of the file after the imports.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.game_of_life_colors
"""

import random
from array import array

import arcade
from arcade import key
from arcade.gl import geometry


CELL_SIZE = 2  # Cell size in pixels
WINDOW_WIDTH = 512  # Width of the window
WINDOW_HEIGHT = 512  # Height of the window
FRAME_DELAY = 2  # The game will only update every 2nd frame


class GameOfLife(arcade.Window):

    def __init__(self, width, height):
        super().__init__(width, height, "Game of Life - Shader Version")
        self.frame = 0

        # Calculate how many cells we need to simulate at this pixel size
        self.texture_size = width // CELL_SIZE, height // CELL_SIZE

        # Create two textures for the next and previous state (RGB textures)
        self.texture_1 = self.ctx.texture(
            self.texture_size,
            components=3,
            filter=(self.ctx.NEAREST, self.ctx.NEAREST),
        )
        self.texture_2 = self.ctx.texture(
            self.texture_size,
            components=3,
            filter=(self.ctx.NEAREST, self.ctx.NEAREST),
        )
        self.write_initial_state()

        # Add the textures to framebuffers so we can render to them
        self.fbo_1 = self.ctx.framebuffer(color_attachments=[self.texture_1])
        self.fbo_2 = self.ctx.framebuffer(color_attachments=[self.texture_2])

        # Fullscreen quad (using triangle strip)
        self.quad_fs = geometry.quad_2d_fs()

        # Shader to draw the texture
        self.display_program = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_vert;
            in vec2 in_uv;
            out vec2 uv;

            void main() {
                gl_Position = vec4(in_vert, 0.0, 1.0);
                uv = in_uv;
            }
            """,
            fragment_shader="""
            #version 330

            uniform sampler2D texture0;
            out vec4 fragColor;
            in vec2 uv;

            void main() {
                fragColor = texture(texture0, uv);
            }
            """,
        )

        # Shader which calculates the next game state.
        # It uses the previous state as input (texture0) and
        # renders the next state into the second texture.
        self.life_program = self.ctx.program(
            vertex_shader="""
            #version 330
            in vec2 in_vert;

            void main() {
                gl_Position = vec4(in_vert, 0.0, 1.0);
            }
            """,
            fragment_shader="""
            #version 330

            uniform sampler2D texture0;
            out vec4 fragColor;

            // Check if something is living in the cell
            bool cell(vec4 fragment) {
                return length(fragment.xyz) > 0.1;
            }

            void main() {
                // Get the pixel position we are currently writing
                ivec2 pos = ivec2(gl_FragCoord.xy);

                // Grab neighbor fragments + current one
                vec4 v1 = texelFetch(texture0, pos + ivec2(-1, -1), 0);
                vec4 v2 = texelFetch(texture0, pos + ivec2( 0, -1), 0);
                vec4 v3 = texelFetch(texture0, pos + ivec2( 1, -1), 0);

                vec4 v4 = texelFetch(texture0, pos + ivec2(-1, 0), 0);
                vec4 v5 = texelFetch(texture0, pos, 0);
                vec4 v6 = texelFetch(texture0, pos + ivec2(1,  0), 0);

                vec4 v7 = texelFetch(texture0, pos + ivec2(-1, 1), 0);
                vec4 v8 = texelFetch(texture0, pos + ivec2( 0, 1), 0);
                vec4 v9 = texelFetch(texture0, pos + ivec2( 1, 1), 0);

                // Cell in current position is alive?
                bool living = cell(v5);

                // Count how many neighbors is alive
                int neighbors = 0;
                if (cell(v1)) neighbors++;
                if (cell(v2)) neighbors++;
                if (cell(v3)) neighbors++;
                if (cell(v4)) neighbors++;
                if (cell(v6)) neighbors++;
                if (cell(v7)) neighbors++;
                if (cell(v8)) neighbors++;
                if (cell(v9)) neighbors++;

                // Average color for all neighbors
                vec4 sum = (v1 + v2 + v3 + v4 + v6 + v7 + v8 + v9) / float(neighbors);

                if (living) {
                    if (neighbors == 2 || neighbors == 3) {
                        // The cell lives, but we write out the average color minus a small value
                        fragColor = vec4(sum.rgb - vec3(1.0/255.0), 1.0);
                    } else {
                        // The cell dies when too few or too many neighbors
                        fragColor = vec4(0.0, 0.0, 0.0, 1.0);
                    }
                } else {
                    if (neighbors == 3) {
                        // A new cell was born
                        fragColor = vec4(normalize(sum.rgb), 1.0);
                    } else {
                        // Still dead
                        fragColor = vec4(0.0, 0.0, 0.0, 1.0);
                    }
                }
            }
            """,
        )

    def gen_initial_data(self, num_values: int):
        """Generate initial data.

        We need to be careful about the initial game state. Carelessly
        random numbers will make the simulation die in only a few frames.
        Instead, we need to generate values which leave room for life
        to exist.

        The implementtation below is one of the slowest possible ways to
        would generate the initial data, but it keeps things simple for
        this example.
        """
        choices = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 32, 64, 128, 192, 255]
        for i in range(num_values):
            yield random.choice(choices)

    def write_initial_state(self):
        """Write initial data to the source texture."""
        size = self.texture_size
        self.texture_1.write(array("B", self.gen_initial_data(size[0] * size[1] * 3)))

    def on_draw(self):
        self.clear()

        # Should we do an update this frame?
        if self.frame % FRAME_DELAY == 0:
            # Calculate the next state
            self.fbo_2.use()  # Render to texture 2
            self.texture_1.use()  # Take texture 1 as input
            self.quad_fs.render(self.life_program)  # Run the life program

            # Draw result to screen
            self.ctx.screen.use()  # Switch back to rendering to screen
            self.texture_2.use()  # Take texture 2 as input
            self.quad_fs.render(self.display_program)  # Display the texture

            # Swap things around for the next frame
            self.texture_1, self.texture_2 = self.texture_2, self.texture_1
            self.fbo_1, self.fbo_2 = self.fbo_2, self.fbo_1
        # Otherwise just draw the current texture
        else:
            # Draw the current texture to the screen
            self.ctx.screen.use()  # Switch back to rendering to screen
            self.texture_1.use()  # Take texture 2 as input
            self.quad_fs.render(self.display_program)  # Display the texture

    def on_update(self, delta_time: float):
        # Track the number of frames
        self.frame += 1

    def on_key_press(self, symbol: int, modifiers: int):
        if symbol == key.SPACE:
            self.write_initial_state()


if __name__ == "__main__":
    game = GameOfLife(WINDOW_WIDTH, WINDOW_HEIGHT)
    game.run()



"""
Display random animal facts from
* https://github.com/wh-iterabb-it/meowfacts
* https://kinduff.github.io/dog-api/

and random cat images from
* https://cataas.com
* https://random.dog

More free APIs: https://github.com/public-apis/public-apis

This example shows how to use multiprocessing to run a service in a separate process.
The service will run in the background and return data to the main process when it's ready.
This is a good way to keep your main game loop running smoothly while waiting for data.
This method is acceptable for simpler types of games. Some extra latency may be introduced
when using this method.

The example also shows that we can send textures between processes.

Some of these services may go down or change their API, so this code may not work
as expected. If you find a service that is no longer working, please make a pull request!

Controls:
    SPACE: Request a new fact OR wait for the next fact to be displayed
    F: Toggle fullscreen
    ESC: Close the window

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.net_process_animal_facts
"""
import PIL.Image
import random
import traceback
import time
import json
import urllib.request
import multiprocessing as mp
from queue import Empty
import arcade
from arcade.math import clamp

CHOICES = ["meow", "woof"]  # Which facts to display


class GameView(arcade.View):
    """Display a random cat fact"""
    def __init__(self):
        super().__init__()
        self.loading = False
        self.text_fact = arcade.Text(
            "",
            x=self.window.width / 2, y=self.window.height / 2 + 50,
            width=1100, font_size=36, anchor_x="center", anchor_y="center",
            multiline=True,
        )
        self.text_info = arcade.Text(
            "Press SPACE to request new fact",
            font_size=20,
            x=20, y=40, color=arcade.color.LIGHT_BLUE,
        )
        self.angle = 0  # Rotation for the spinning loading initicator
        # Keep track of time to auto request new facts
        self.update_interval = 30  # seconds
        self.last_updated = 0
        # Background image
        self.bg_texture_1 = None
        self.bg_texture_2 = None
        # Keep track of when we last updated the background image (for fading)
        self.bg_updated_time = 0

    def on_draw(self):
        """Draw the next frame"""
        self.clear()

        # Draw a background image fading between two images if needed
        fade = clamp(time.time() - self.bg_updated_time, 0.0, 1.0)
        self.draw_background(self.bg_texture_1, fade * 0.35)
        if fade < 1.0:
            self.draw_background(self.bg_texture_2,  0.35 * (1.0 - fade))

        # Draw the fact text
        self.text_fact.draw()

        # Draw a spinning circle while we wait for a fact
        if self.loading:
            arcade.draw_rect_filled(
                arcade.XYWH(self.window.width - 50, self.window.height - 50, 50, 50),
                color=arcade.color.ORANGE,
                tilt_angle=self.angle,
            )

        # Draw a progress bar to show how long until we request a new fact
        progress = 1.0 - ((time.time() - self.last_updated) / self.update_interval)
        if progress > 0:
            arcade.draw_lrbt_rectangle_filled(
                left=0, right=self.window.width * progress,
                bottom=0, top=20,
                color=arcade.color.LIGHT_KHAKI
            )

        # Draw info text
        self.text_info.draw()

    def draw_background(self, texture: arcade.Texture, alpha: int) -> None:
        """Draw a background image attempting to fill the entire window"""
        if texture is None:
            return

        # Get the highest ratio of width or height to fill the window
        scale = max(self.window.width / texture.width, self.window.height / texture.height)
        arcade.draw_texture_rect(
            texture,
            arcade.XYWH(
                self.window.width / 2,
                self.window.height / 2,
                texture.width * scale,
                texture.height * scale,
            ),
            blend=True,
            alpha=int(alpha * 255),
        )

    def on_update(self, delta_time: float):
        """Update state for the next frame"""
        # Keep spinning the loading indicator
        self.angle += delta_time * 400

        # Request a new fact if it's time to do so
        if time.time() - self.last_updated > self.update_interval:
            self.request_new_fact()

        # Check if we have some new data from the service
        # This is a non-blocking call, so it will return immediately (fast)
        try:
            # Immediately raises Empty exception if there is no data
            data = animal_service.out_queue.get(block=False)
            # If we got a string, it's new fact text
            if isinstance(data, str):
                self.text_fact.text = data
                self.loading = False
                self.last_updated = time.time()
            # If we got a texture, it's a new background image
            elif isinstance(data, arcade.Texture):
                self.bg_texture_2 = self.bg_texture_1
                self.bg_texture_1 = data
                self.bg_updated_time = time.time()
        except Empty:
            pass

    def on_key_press(self, symbol: int, modifiers: int):
        """Handle key presses"""
        if symbol == arcade.key.SPACE:
            self.request_new_fact()
        elif symbol == arcade.key.F:
            self.window.set_fullscreen(not self.window.fullscreen)
            self.window.set_vsync(True)
        elif symbol == arcade.key.ESCAPE:
            self.window.close()

    def request_new_fact(self):
        """Request a new fact from the service unless we're already waiting for one"""
        if not self.loading:
            self.loading = True
            animal_service.in_queue.put(random.choice(CHOICES))

    def on_resize(self, width: float, height: float):
        """Handle window resize events"""
        # Re-position the text to the center of the window
        # and make sure it fits in the window (width)
        self.text_fact.x = width / 2
        self.text_fact.y = height / 2 + 50
        self.text_fact.width = width - 200


class AnimaFactsService:
    """
    A daemon process that will keep running and return random cat facts.
    The process will die with the main process.
    """
    def __init__(self):
        # Queue for sending commands to the process
        self.in_queue = mp.Queue()
        # Queue for receiving data from the process
        self.out_queue = mp.Queue()
        # The process itself. Daemon=True means it will die with the main process.
        self.process = mp.Process(
            target=self.do_work,
            args=(self.in_queue, self.out_queue),
            daemon=True,
        )

    def do_work(self, in_queue, out_queue):
        """Keep pulling the in queue forever and do work when something arrives"""
        fact_types = {
            "meow": CatFacts(),
            "woof": DogFacts(),
        }
        # Keep looping forever waiting for work
        while True:
            # The supported commands are "ready", "meow", and "woof"
            command = in_queue.get(block=True)
            if command == "ready":
                # Signal that the process is ready to receive commands
                out_queue.put("ready")
            else:
                selected_type = fact_types.get(command)
                try:
                    out_queue.put(selected_type.get_fact())
                    out_queue.put(selected_type.get_image())
                except Exception:
                    traceback.print_exc()

    def start(self) -> int:
        """Start the process and wait for it to be ready"""
        self.process.start()
        # Send a ready request and do a blocking get to wait for the response.
        self.in_queue.put("ready")
        self.out_queue.get(block=True)
        return self.process.pid


class Facts:
    """Base class for fact providers"""

    def get_fact(self) -> str:
        raise NotImplementedError()

    def get_image(self) -> arcade.Texture:
        raise NotImplementedError()


class CatFacts(Facts):
    """Get random cat facts and images"""

    def get_fact(self) -> str:
        with urllib.request.urlopen("https://meowfacts.herokuapp.com") as fd:
            data = json.loads(fd.read().decode("utf-8"))
            print(data)
            return data["data"][0]

    def get_image(self) -> arcade.Texture:
        """Get a random cat image from https://cataas.com/"""
        with urllib.request.urlopen("https://cataas.com/cat") as fd:
            return arcade.Texture(PIL.Image.open(fd).convert("RGBA"))


class DogFacts(Facts):
    """Get random dog facts and images"""

    def __init__(self):
        # Fetch the full image list at creation. This is more practical for this api.
        with urllib.request.urlopen("https://random.dog/doggos") as fd:
            self.images = json.loads(fd.read().decode("utf-8"))

        # Remove videos
        self.images = [i for i in self.images if not i.endswith(".mp4")]

    def get_fact(self) -> str:
        with urllib.request.urlopen("https://dogapi.dog/api/v2/facts") as fd:
            data = json.loads(fd.read().decode("utf-8"))
            print(data)
            return data["data"][0]["attributes"]["body"]

    def get_image(self) -> arcade.Texture:
        """Get a random dog image from https://random.dog"""
        name = random.choice(self.images)
        print("Image:", name)
        with urllib.request.urlopen(f"https://random.dog/{name}") as fd:
            return arcade.Texture(PIL.Image.open(fd).convert("RGBA"))


def main():
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(1280, 720, "Random Animal Facts", resizable=True, vsync=True)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()

if __name__ == "__main__":
    """ Main function """
    animal_service = AnimaFactsService()
    print("Starting animal fact service...")
    pid = animal_service.start()
    print("Service started. pid:", pid)

    main()



"""
This simple animation example shows how to use classes to animate
multiple objects on the screen at the same time.

Note: Sprites draw much faster than drawing primitives

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.shapes
"""

import arcade
import random

# Set up the constants
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Shapes!"

NUMBER_OF_SHAPES = 200


class Shape:
    """ Generic base shape class """
    def __init__(self, x, y, width, height, angle, delta_x, delta_y,
                 delta_angle, color):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.angle = angle
        self.delta_x = delta_x
        self.delta_y = delta_y
        self.delta_angle = delta_angle
        self.color = color

    def move(self):
        self.x += self.delta_x
        self.y += self.delta_y
        self.angle += self.delta_angle
        if self.x < 0 and self.delta_x < 0:
            self.delta_x *= -1
        if self.y < 0 and self.delta_y < 0:
            self.delta_y *= -1
        if self.x > WINDOW_WIDTH and self.delta_x > 0:
            self.delta_x *= -1
        if self.y > WINDOW_HEIGHT and self.delta_y > 0:
            self.delta_y *= -1


class Ellipse(Shape):

    def draw(self):
        arcade.draw_ellipse_filled(self.x, self.y, self.width, self.height,
                                   self.color, self.angle)


class Rectangle(Shape):

    def draw(self):
        arcade.draw_rect_filled(arcade.rect.XYWH(self.x, self.y, self.width, self.height),
                                self.color, self.angle)


class Line(Shape):

    def draw(self):
        arcade.draw_line(self.x, self.y,
                         self.x + self.width, self.y + self.height,
                         self.color, 2)


class GameView(arcade.View):
    """ Main application class. """

    def __init__(self):
        # Call the parent __init__
        super().__init__()

        # Create a shape list
        self.shape_list = []

        for i in range(NUMBER_OF_SHAPES):

            # Random spot
            x = random.randrange(0, WINDOW_WIDTH)
            y = random.randrange(0, WINDOW_HEIGHT)

            # Random size
            width = random.randrange(15, 40)
            height = random.randrange(15, 40)

            # Random angle
            angle = random.randrange(0, 360)

            # Random movement
            d_x = random.randrange(-3, 4)
            d_y = random.randrange(-3, 4)
            d_angle = random.randrange(-3, 4)

            # Random color
            red = random.randrange(256)
            green = random.randrange(256)
            blue = random.randrange(256)
            alpha = random.randrange(256)

            # Random line, ellipse, or rect
            shape_type = random.randrange(3)

            if shape_type == 0:
                shape = Rectangle(x, y, width, height, angle, d_x, d_y,
                                  d_angle, (red, green, blue, alpha))
            elif shape_type == 1:
                shape = Ellipse(x, y, width, height, angle, d_x, d_y,
                                d_angle, (red, green, blue, alpha))
            else:
                shape = Line(x, y, width, height, angle, d_x, d_y,
                             d_angle, (red, green, blue, alpha))

            # Add this new shape to the list
            self.shape_list.append(shape)

    def on_update(self, dt):
        """ Move everything """
        for shape in self.shape_list:
            shape.move()

    def on_draw(self):
        """ Render the screen. """
        self.clear()

        # Draw the shapes
        for shape in self.shape_list:
            shape.draw()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



import math

from typing_extensions import override

from arcade import Texture
from arcade.gui import NinePatchTexture
from arcade.gui.property import Property, bind
from arcade.gui.surface import Surface
from arcade.gui.widgets import UIWidget


class UIImage(UIWidget):
    """UIWidget showing a texture.

    If no size given, the texture size is used.

    The UIImage supports rotation and alpha values, which only apply to the texture.
    Border, and background color are not affected by this.
    The size of the image is reduced when rotated to stay within bounce of the widget.

    Args:
        texture: Texture to show
        width: width of widget
        height: height of widget
        angle: angle of the texture in degrees
        alpha: alpha value of the texture, value between 0 and 255
        **kwargs: passed to UIWidget
    """

    texture = Property[Texture | NinePatchTexture]()
    """Texture to show"""
    alpha = Property(255)
    """Alpha value of the texture, value between 0 and 255.
    0 is fully transparent, 255 is fully visible."""
    angle = Property(0)
    """Angle of the texture in degrees.
    The image will be rotated around its center and fitted into the widget size."""

    def __init__(
        self,
        *,
        texture: Texture | NinePatchTexture,
        width: float | None = None,
        height: float | None = None,
        angle: int = 0,
        alpha: int = 255,
        **kwargs,
    ):
        self.texture = texture
        self.angle = angle
        self.alpha = alpha

        super().__init__(
            width=width if width else texture.width,
            height=height if height else texture.height,
            **kwargs,
        )
        bind(self, "texture", self.trigger_render)
        bind(self, "alpha", self.trigger_full_render)
        bind(self, "angle", self.trigger_full_render)

    @override
    def do_render(self, surface: Surface):
        """Render the stored texture in the size of the widget."""
        if self.texture:
            self.prepare_render(surface)

            if self.angle == 0:
                surface.draw_texture(
                    x=0,
                    y=0,
                    width=self.content_width,
                    height=self.content_height,
                    tex=self.texture,
                    alpha=self.alpha,
                )
            else:
                w = self.content_width
                h = self.content_height
                angle_radians = math.radians(self.angle)
                cos_value = abs(math.cos(angle_radians))
                sin_value = abs(math.sin(angle_radians))

                # https://stackoverflow.com/a/33867165/2947505
                # Calculate the minimum size of the rotated image
                # W = w·|cos φ| + h·|sin φ|
                w_rotated = w * cos_value + h * sin_value
                # H = w·|sin φ| + h·|cos φ|
                h_rotated = w * sin_value + h * cos_value
                # a = min(wo / W, ho / H)
                factor = min(w / w_rotated, h / h_rotated)
                # W′ = a·w
                w_fitted = factor * w
                # H′ = a·h
                h_fitted = factor * h

                draw_rect = self.content_rect.align_left(0).align_bottom(0)
                draw_rect = draw_rect.resize(w_fitted, h_fitted)
                surface.draw_texture(
                    x=draw_rect.left,
                    y=draw_rect.bottom,
                    width=draw_rect.width,
                    height=draw_rect.height,
                    tex=self.texture,
                    alpha=self.alpha,
                    angle=self.angle,
                )



from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from typing import TYPE_CHECKING

from pyglet.math import Mat4, Vec2, Vec3
from typing_extensions import Self

from arcade.types import LBWH, Point, Rect
from arcade.window_commands import get_window

if TYPE_CHECKING:
    from arcade.context import ArcadeContext

__all__ = ["ViewportProjector", "DefaultProjector"]


class ViewportProjector:
    """
    A simple Projector which does not rely on any camera PoDs.

    Does not have a way of moving, rotating, or zooming the camera.
    perfect for something like UI or for mapping to an offscreen framebuffer.

    Args:
        viewport: The viewport to project to.
        context: The window context to bind the camera to. Defaults to the currently active window.
    """

    def __init__(
        self,
        viewport: Rect | None = None,
        *,
        context: ArcadeContext | None = None,
    ):
        self._ctx: ArcadeContext = context or get_window().ctx
        self._viewport: Rect = viewport or LBWH(*self._ctx.viewport)
        self._projection_matrix: Mat4 = Mat4.orthogonal_projection(
            0.0, self._viewport.width, 0.0, self._viewport.height, -100, 100
        )

    @property
    def viewport(self) -> Rect:
        """
        The viewport use to derive projection and view matrix.
        """
        return self._viewport

    @viewport.setter
    def viewport(self, viewport: Rect) -> None:
        self._viewport = viewport
        self._projection_matrix = Mat4.orthogonal_projection(
            0, viewport.width, 0, viewport.height, -100, 100
        )

    def use(self) -> None:
        """
        Set the window's projection and view matrix.
        Also sets the projector as the windows current camera.
        """
        self._ctx.current_camera = self

        self._ctx.viewport = self.viewport.lbwh_int  # get the integer 4-tuple LBWH

        self._ctx.view_matrix = Mat4()
        self._ctx.projection_matrix = self._projection_matrix

    @contextmanager
    def activate(self) -> Generator[Self, None, None]:
        """
        The context manager version of the use method.

        usable with the 'with' block. e.g. 'with ViewportProjector.activate() as cam: ...'
        """
        previous = self._ctx.current_camera
        try:
            self.use()
            yield self
        finally:
            previous.use()

    def project(self, world_coordinate: Point) -> Vec2:
        """
        Take a Vec2 or Vec3 of coordinates and return the related screen coordinate
        """
        x, y, *z = world_coordinate
        return Vec2(x, y)

    def unproject(self, screen_coordinate: Point) -> Vec3:
        """
        Map the screen pos to screen_coordinates.

        Due to the nature of viewport projector this does not do anything.
        """
        x, y, *_z = screen_coordinate
        z = 0.0 if not _z else _z[0]

        return Vec3(x, y, z)


# As this class is only supposed to be used internally
# I wanted to place an _ in front, but the linting complains
# about it being a protected class.
class DefaultProjector(ViewportProjector):
    """
    An extremely limited projector which lacks any kind of control. This is only
    here to act as the default camera used internally by Arcade. There should be
    no instance where a developer would want to use this class.

    Args:
        context: The window context to bind the camera to. Defaults to the currently active window.
    """

    def __init__(self, *, context: ArcadeContext | None = None):
        super().__init__(context=context)

    def use(self) -> None:
        """
        Set the window's Projection and View matrices.

        cache's the window viewport to determine the projection matrix.
        """

        viewport = self.viewport.lbwh_int
        # If the viewport is correct and the default camera is in use,
        # then don't waste time resetting the view and projection matrices
        if self._ctx.viewport == viewport and self._ctx.current_camera == self:
            return

        # If the viewport has changed while the default camera is active then the
        # default needs to update itself.
        # If it was another camera's viewport being used the default camera should not update.
        if self._ctx.viewport != viewport and self._ctx.current_camera == self:
            self.viewport = LBWH(*self._ctx.viewport)
        else:
            self._ctx.viewport = viewport

        self._ctx.current_camera = self

        self._ctx.view_matrix = Mat4()
        self._ctx.projection_matrix = self._projection_matrix



from . import video
from . import light
from . import input
from . import background
from . import splash

from .texture_render_target import RenderTargetTexture

__all__ = ["video", "light", "input", "background", "RenderTargetTexture", "splash"]



"""Create custom scalable UI themes with NinePatchTexture

Nine-patch textures are a technique for scalable custom borders and
frames for UI elements. Widgets which support a background texture can
also use a NinePatchTexture to support scaling where the corners stay the
same, like a window.

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.ninepatch
"""

import arcade
from arcade import load_texture
from arcade.gui import UIManager, UIAnchorLayout, UIWidget, NinePatchTexture


class MyView(arcade.View):
    def __init__(self):
        super().__init__()
        self.ui = UIManager()

        # Setup widget and use background with ninepatch information
        self.nine_patch_widget = UIWidget(size_hint=(0.5, 0.5))
        self.nine_patch_widget.with_background(
            texture=NinePatchTexture(
                texture=load_texture(":resources:gui_basic_assets/window/grey_panel.png"),
                left=7,
                right=7,
                bottom=7,
                top=7,
            )
        )

        self.ui.add(UIAnchorLayout(children=[self.nine_patch_widget]))

    def on_show_view(self):
        self.window.background_color = arcade.color.DARK_BLUE_GRAY
        # Enable UIManager when view is shown to catch window events
        self.ui.enable()

    def on_hide_view(self):
        self.ui.disable()

    def on_draw(self):
        self.clear()
        self.ui.draw()


def main():
    window = arcade.Window(800, 600, "UIExample", resizable=True)
    window.show_view(MyView())
    window.run()


if __name__ == "__main__":
    main()



import array

from arcade import gl
from arcade.types import Color, Point2, Point2List, RGBOrA255
from arcade.window_commands import get_window

from .helpers import _generic_draw_line_strip, get_points_for_thick_line


def draw_line_strip(point_list: Point2List, color: RGBOrA255, line_width: float = 1) -> None:
    """
    Draw a multi-point line.

    Args:
        point_list:
            List of x, y points that make up this strip
        color:
            A color, specified as an RGBA tuple or a
            :py:class:`.Color` instance.
        line_width:
            Width of the line
    """
    if line_width == 1:
        _generic_draw_line_strip(point_list, color, gl.LINE_STRIP)
    else:
        triangle_point_list: list[Point2] = []
        # FIXME: This needs a lot of improvement
        last_point = None
        for point in point_list:
            if last_point is not None:
                points = get_points_for_thick_line(
                    last_point[0], last_point[1], point[0], point[1], line_width
                )
                reordered_points = points[1], points[0], points[2], points[3]
                triangle_point_list.extend(reordered_points)
            last_point = point
        _generic_draw_line_strip(triangle_point_list, color, gl.TRIANGLE_STRIP)


def draw_line(
    start_x: float,
    start_y: float,
    end_x: float,
    end_y: float,
    color: RGBOrA255,
    line_width: float = 1,
) -> None:
    """
    Draw a line.

    Args:
        start_x:
            x position of line starting point.
        start_y:
            y position of line starting point.
        end_x:
            x position of line ending point.
        end_y:
            y position of line ending point.
        color:
            A color, specified as an RGBA tuple or a
            :py:class:`.Color` instance.
        line_width:
            Width of the line in pixels.
    """
    # Fail if we don't have a window, context, or right GL abstractions
    window = get_window()
    ctx = window.ctx
    program = ctx.shape_line_program
    geometry = ctx.shape_line_geometry  # type: ignore
    line_pos_buffer = ctx.shape_line_buffer_pos  # type: ignore

    # Validate & normalize to a pass the shader an RGBA float uniform
    color_normalized = Color.from_iterable(color).normalized

    # Pass data to the shader
    program["color"] = color_normalized
    program["line_width"] = line_width
    line_pos_buffer.orphan()  # Allocate new buffer internally
    line_pos_buffer.write(data=array.array("f", (start_x, start_y, end_x, end_y)))

    ctx.enable(ctx.BLEND)
    geometry.render(program, mode=gl.LINES, vertices=2)
    ctx.disable(ctx.BLEND)


def draw_lines(point_list: Point2List, color: RGBOrA255, line_width: float = 1) -> None:
    """
    Draw a set of lines.

    Draw a line between each pair of points specified.

    Args:
        point_list:
            List of points making up the lines. Each point is in a list.
            So it is a list of lists.
        color:
            A color, specified as an RGBA tuple or a
            :py:class:`.Color` instance.
        line_width:
            Width of the line in pixels.
    """
    # Fail if we don't have a window, context, or right GL abstractions
    window = get_window()
    ctx = window.ctx
    program = ctx.shape_line_program
    geometry = ctx.shape_line_geometry  # type: ignore
    line_buffer_pos = ctx.shape_line_buffer_pos  # type: ignore

    # Validate & normalize to a pass the shader an RGBA float uniform
    color_normalized = Color.from_iterable(color).normalized

    line_pos_array = array.array("f", (v for point in point_list for v in point))
    num_points = len(point_list)

    # Grow buffer until large enough to hold all our data
    goal_buffer_size = num_points * 3 * 4
    while goal_buffer_size > line_buffer_pos.size:
        ctx.shape_line_buffer_pos.orphan(line_buffer_pos.size * 2)  # type: ignore
    else:
        ctx.shape_line_buffer_pos.orphan()  # type: ignore

    ctx.enable(ctx.BLEND)

    # Pass data to shader
    program["line_width"] = line_width
    program["color"] = color_normalized
    line_buffer_pos.write(data=line_pos_array)

    geometry.render(program, mode=gl.LINES, vertices=num_points)

    ctx.disable(ctx.BLEND)



"""
Conway's Game of Life

This code shows how to set up sprites in a grid, and then use their
'alpha' value to quickly turn them on and off.

After installing the `arcade` package version 2.4.4+, this program can be run by
typing:
python -m arcade.examples.conway_alpha
"""
import arcade
import random

# Set how many rows and columns we will have
ROW_COUNT = 70
COLUMN_COUNT = 128

# This sets the WIDTH and HEIGHT of each grid location
CELL_WIDTH = 15
CELL_HEIGHT = 15

# This sets the margin between each cell
# and on the edges of the screen.
CELL_MARGIN = 0

# Do the math to figure out our screen dimensions
WINDOW_WIDTH = (CELL_WIDTH + CELL_MARGIN) * COLUMN_COUNT + CELL_MARGIN
WINDOW_HEIGHT = (CELL_HEIGHT + CELL_MARGIN) * ROW_COUNT + CELL_MARGIN
WINDOW_TITLE = "Conway's Game of Life"

# Colors and alpha values
ALIVE_COLOR = arcade.color.BISTRE
BACKGROUND_COLOR = arcade.color.ANTIQUE_WHITE
ALPHA_ON = 255
ALPHA_OFF = 0


def create_grids():
    """
    Create a 2D and 1D grid of sprites. We use the 1D SpriteList for drawing,
    and the 2D list for accessing via grid. Both lists point to the same set of
    sprites.
    """
    # One dimensional list of all sprites in the two-dimensional sprite list
    grid_sprites_one_dim = arcade.SpriteList()

    # This will be a two-dimensional grid of sprites to mirror the two
    # dimensional grid of numbers. This points to the SAME sprites that are
    # in grid_sprite_list, just in a 2d manner.
    grid_sprites_two_dim = []

    # Create a list of sprites to represent each grid location
    for row in range(ROW_COUNT):
        grid_sprites_two_dim.append([])

        for column in range(COLUMN_COUNT):

            # Make the sprite as a soft circle
            sprite = arcade.SpriteCircle(CELL_WIDTH // 2, ALIVE_COLOR, soft=True)

            # Position the sprite
            x = column * (CELL_WIDTH + CELL_MARGIN) + (CELL_WIDTH / 2 + CELL_MARGIN)
            y = row * (CELL_HEIGHT + CELL_MARGIN) + (CELL_HEIGHT / 2 + CELL_MARGIN)
            sprite.center_x = x
            sprite.center_y = y

            # Add the sprite to both lists
            grid_sprites_one_dim.append(sprite)
            grid_sprites_two_dim[row].append(sprite)

    return grid_sprites_one_dim, grid_sprites_two_dim


def randomize_grid(grid: arcade.SpriteList):
    """ Randomize the grid to alive/dead """
    for cell in grid:
        pick = random.randrange(2)
        if pick:
            cell.alpha = ALPHA_ON
        else:
            cell.alpha = ALPHA_OFF


class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """
        Set up the application.
        """
        super().__init__()

        self.background_color = BACKGROUND_COLOR

        # We need two layers. One holds the current state of our grid, the other
        # holds the next frame's state. We flip back and forth between the two.
        grid_sprites_one_dim1, grid_sprites_two_dim1 = create_grids()
        grid_sprites_one_dim2, grid_sprites_two_dim2 = create_grids()

        self.layers_grid_sprites_one_dim = [grid_sprites_one_dim1, grid_sprites_one_dim2]
        self.layers_grid_sprites_two_dim = [grid_sprites_two_dim1, grid_sprites_two_dim2]

        self.cur_layer = 0
        randomize_grid(self.layers_grid_sprites_one_dim[0])

    def reset(self):
        """ Reset the grid """
        randomize_grid(self.layers_grid_sprites_one_dim[0])

    def on_draw(self):
        """ Render the screen. """
        # Clear all pixels in the window
        self.clear()
        self.layers_grid_sprites_one_dim[0].draw()

    def on_key_press(self, symbol: int, modifiers: int):
        """ Handle key press events """
        if symbol == arcade.key.SPACE:
            self.reset()
        elif symbol == arcade.key.ESCAPE:
            self.window.close()

    def on_update(self, delta_time: float):
        """ Update the grid """

        # Flip layers
        if self.cur_layer == 0:
            layer1 = self.layers_grid_sprites_two_dim[0]
            layer2 = self.layers_grid_sprites_two_dim[1]
            self.cur_layer = 1
        else:
            layer1 = self.layers_grid_sprites_two_dim[1]
            layer2 = self.layers_grid_sprites_two_dim[0]
            self.cur_layer = 0

        # Count the neighbors that are alive
        for row in range(ROW_COUNT):
            for column in range(COLUMN_COUNT):
                live_neighbors = 0
                # -1 -1
                if row > 0 and column > 0 \
                        and layer1[row - 1][column - 1].alpha == ALPHA_ON:
                    live_neighbors += 1
                # -1  0
                if row > 0 and layer1[row - 1][column].alpha == ALPHA_ON:
                    live_neighbors += 1
                # -1 +1
                if row > 0 and column < COLUMN_COUNT - 1\
                        and layer1[row - 1][column + 1].alpha == ALPHA_ON:
                    live_neighbors += 1
                #  0 +1
                if column < COLUMN_COUNT - 1 \
                        and layer1[row][column + 1].alpha == ALPHA_ON:
                    live_neighbors += 1
                # +1 +1
                if row < ROW_COUNT - 1 \
                        and column < COLUMN_COUNT - 1 \
                        and layer1[row + 1][column + 1].alpha == ALPHA_ON:
                    live_neighbors += 1
                # +1  0
                if row < ROW_COUNT - 1 and layer1[row + 1][column].alpha == ALPHA_ON:
                    live_neighbors += 1
                # +1 -1
                if row < ROW_COUNT - 1 and column > 0 \
                        and layer1[row + 1][column - 1].alpha == ALPHA_ON:
                    live_neighbors += 1
                #  0 -1
                if column > 0 and layer1[row][column - 1].alpha == ALPHA_ON:
                    live_neighbors += 1

                """
                Implement Conway's game of life rules

                Any live cell with two or three live neighbours survives.
                Any dead cell with three live neighbours becomes a live cell.
                All other live cells die in the next generation. Similarly,
                all other dead cells stay dead.
                """
                # Shortcut the cell sprites
                l1_sprite = layer1[row][column]
                l2_sprite = layer2[row][column]

                if l1_sprite.alpha == ALPHA_ON and (live_neighbors == 2 or live_neighbors == 3):
                    if l2_sprite.alpha == ALPHA_OFF:
                        l2_sprite.alpha = ALPHA_ON
                elif l1_sprite.alpha == ALPHA_OFF and live_neighbors == 3:
                    if l2_sprite.alpha == ALPHA_OFF:
                        l2_sprite.alpha = ALPHA_ON
                else:
                    if l2_sprite.alpha == ALPHA_ON:
                        l2_sprite.alpha = ALPHA_OFF


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    window.center_window()

    # Create and setup the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



from pyglet import gl

# Texture min/mag filters
NEAREST = 0x2600
LINEAR = 0x2601
NEAREST_MIPMAP_NEAREST = 0x2700
LINEAR_MIPMAP_NEAREST = 0x2701
NEAREST_MIPMAP_LINEAR = 0x2702
LINEAR_MIPMAP_LINEAR = 0x2703

# Texture wrapping
REPEAT = gl.GL_REPEAT
CLAMP_TO_EDGE = gl.GL_CLAMP_TO_EDGE
CLAMP_TO_BORDER = gl.GL_CLAMP_TO_BORDER
MIRRORED_REPEAT = gl.GL_MIRRORED_REPEAT

# Blend functions
ZERO = 0x0000
ONE = 0x0001
SRC_COLOR = 0x0300
ONE_MINUS_SRC_COLOR = 0x0301
SRC_ALPHA = 0x0302
ONE_MINUS_SRC_ALPHA = 0x0303
DST_ALPHA = 0x0304
ONE_MINUS_DST_ALPHA = 0x0305
DST_COLOR = 0x0306
ONE_MINUS_DST_COLOR = 0x0307

# Blend equations

FUNC_ADD = 0x8006
"""source + destination"""

FUNC_SUBTRACT = 0x800A
"""source - destination"""

FUNC_REVERSE_SUBTRACT = 0x800B
"""destination - source"""

MIN = 0x8007
"""Minimum of source and destination"""

MAX = 0x8008
"""Maximum of source and destination"""

BLEND_DEFAULT = SRC_ALPHA, ONE_MINUS_SRC_ALPHA
BLEND_ADDITIVE = ONE, ONE
BLEND_PREMULTIPLIED_ALPHA = SRC_ALPHA, ONE

# VertexArray: Primitives
POINTS = 0
LINES = 1
LINE_LOOP = 2
LINE_STRIP = 3
TRIANGLES = 4
TRIANGLE_STRIP = 5
TRIANGLE_FAN = 6
LINES_ADJACENCY = 10
LINE_STRIP_ADJACENCY = 11
TRIANGLES_ADJACENCY = 12
TRIANGLE_STRIP_ADJACENCY = 13
PATCHES = 14
